"use strict";

exports.__esModule = true;
exports.delIns = delIns;
exports.delInsM = delInsM;
exports.delStaticInsM = delStaticInsM;
exports.makeWaKey = makeWaKey;
exports.mapIns = mapIns;
exports.mapInsM = mapInsM;
exports.mapStaticIns = mapStaticIns;
exports.mapStaticInsM = mapStaticInsM;
exports.mayRecoverDepRecord = mayRecoverDepRecord;
exports.waKey2uKeyMap = exports.waKey2staticUKeyMap = void 0;

var _util = require("../support/util");

/* eslint-disable camelcase */
// 依赖收集写入的映射
var waKey2uKeyMap = {}; // 依赖标记写入的映射，是一个实例化完成就会固化的依赖
// 不采取一开始映射好全部waKey的形式，而是采用safeGet动态添加map映射

exports.waKey2uKeyMap = waKey2uKeyMap;
var waKey2staticUKeyMap = {};
exports.waKey2staticUKeyMap = waKey2staticUKeyMap;

function _mapIns(mapContainer, waKey, ccUniqueKey) {
  try {
    mapContainer[waKey][ccUniqueKey] = 1; //处于依赖状态
  } catch (err) {
    var map = {};
    map[ccUniqueKey] = 1;
    mapContainer[waKey] = map;
  }
}

function makeWaKey(module, stateKey) {
  return module + "/" + stateKey;
}

function mapIns(module, stateKey, ccUniqueKey) {
  _mapIns(waKey2uKeyMap, makeWaKey(module, stateKey), ccUniqueKey);
}

function mapInsM(modStateKey, ccUniqueKey) {
  _mapIns(waKey2uKeyMap, modStateKey, ccUniqueKey);
}

function delIns(module, stateKey, ccUniqueKey) {
  delInsM(makeWaKey(module, stateKey), ccUniqueKey);
}

var cUkey2depKeys = {};

function delInsM(modStateKey, ccUniqueKey) {
  try {
    // 对抗 react 18 里的 strict 双调用模型
    if ((0, _util.shouldGuessStrictMode)()) {
      (0, _util.safeGetThenNoDupPush)(cUkey2depKeys, ccUniqueKey, modStateKey);
    }

    delete waKey2uKeyMap[modStateKey][ccUniqueKey];
  } catch (err) {// do nothing
  }
}

function mapStaticIns(module, stateKey, ccUniqueKey) {
  _mapIns(waKey2staticUKeyMap, makeWaKey(module, stateKey), ccUniqueKey);
}

function mapStaticInsM(modStateKey, ccUniqueKey) {
  _mapIns(waKey2staticUKeyMap, modStateKey, ccUniqueKey);
}

var cUkey2staticDepKeys = {};

function delStaticInsM(modStateKey, ccUniqueKey) {
  try {
    // 对抗 react 18 里的 strict 双调用模型
    if ((0, _util.shouldGuessStrictMode)()) {
      (0, _util.safeGetThenNoDupPush)(cUkey2staticDepKeys, ccUniqueKey, modStateKey);
    }

    delete waKey2staticUKeyMap[modStateKey][ccUniqueKey];
  } catch (err) {// do nothing
  }
}

function mayRecoverDepRecord(ccUniqueKey) {
  if ((0, _util.shouldGuessStrictMode)()) {
    var depKeys = cUkey2depKeys[ccUniqueKey]; // 恢复时顺带清理掉

    if (depKeys) {
      delete cUkey2depKeys[ccUniqueKey];
      depKeys.forEach(function (key) {
        return mapInsM(key, ccUniqueKey);
      });
    }

    var staticDepKeys = cUkey2staticDepKeys[ccUniqueKey] || [];

    if (staticDepKeys) {
      delete cUkey2staticDepKeys[ccUniqueKey];
      staticDepKeys.forEach(function (key) {
        return mapStaticInsM(key, ccUniqueKey);
      });
    }
  }
}