"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

exports.__esModule = true;
exports.configMiddlewares = configMiddlewares;
exports.configPlugins = configPlugins;
exports.configRootComputed = configRootComputed;
exports.configRootLifecycle = configRootLifecycle;
exports.configRootReducer = configRootReducer;
exports.configRootWatch = configRootWatch;
exports.configStoreState = configStoreState;
exports["default"] = void 0;

var util = _interopRequireWildcard(require("../../support/util"));

var _privConstant = require("../../support/priv-constant");

var _constant = require("../../support/constant");

var _ccContext = _interopRequireDefault(require("../../cc-context"));

var _initModuleState = _interopRequireDefault(require("../state/init-module-state"));

var _initModuleReducer = _interopRequireDefault(require("../reducer/init-module-reducer"));

var _initModuleWatch = _interopRequireDefault(require("../watch/init-module-watch"));

var _initModuleComputed = _interopRequireDefault(require("../computed/init-module-computed"));

var _initModuleLifecycle = _interopRequireDefault(require("./init-module-lifecycle"));

var _plugin = require("../plugin");

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function _getRequireWildcardCache(nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { "default": obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj["default"] = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

var isPJO = util.isPJO,
    okeys = util.okeys,
    isObject = util.isObject,
    evalState = util.evalState;

function checkObj(rootObj, tag) {
  if (!isPJO(rootObj)) {
    throw new Error(tag + " " + _privConstant.INAJ);
  }
}

function configStoreState(storeState) {
  checkObj(storeState, 'state');
  delete storeState[_constant.MODULE_VOID];
  delete storeState[_constant.MODULE_CC];
  if (!storeState[_constant.MODULE_GLOBAL]) storeState[_constant.MODULE_GLOBAL] = {};
  if (!storeState[_constant.MODULE_DEFAULT]) storeState[_constant.MODULE_DEFAULT] = {};
  var moduleNames = okeys(storeState);
  var len = moduleNames.length;

  for (var i = 0; i < len; i++) {
    var moduleName = moduleNames[i];
    var moduleState = evalState(storeState[moduleName], "module[" + moduleName + "] ");
    (0, _initModuleState["default"])(moduleName, moduleState);
  }
}
/**
 * @param {{[moduleName:string]:{[reducerFnType:string]:function}}} rootReducer 
 */


function configRootReducer(rootReducer, rootGhost) {
  checkObj(rootReducer, 'reducer');
  if (!isObject(rootReducer[_constant.MODULE_DEFAULT])) rootReducer[_constant.MODULE_DEFAULT] = {};
  if (!isObject(rootReducer[_constant.MODULE_GLOBAL])) rootReducer[_constant.MODULE_GLOBAL] = {};
  okeys(rootReducer).forEach(function (m) {
    return (0, _initModuleReducer["default"])(m, rootReducer[m], rootGhost[m]);
  });
}

function configRootComputed(rootComputed) {
  checkObj(rootComputed, 'computed');
  okeys(rootComputed).forEach(function (m) {
    return (0, _initModuleComputed["default"])(m, rootComputed[m]);
  });
}

function configRootWatch(rootWatch) {
  checkObj(rootWatch, 'watch');
  okeys(rootWatch).forEach(function (m) {
    return (0, _initModuleWatch["default"])(m, rootWatch[m]);
  });
}

function configRootLifecycle(rootLifecycle) {
  checkObj(rootLifecycle, 'lifecycle');
  okeys(rootLifecycle).forEach(function (m) {
    return (0, _initModuleLifecycle["default"])(m, rootLifecycle[m]);
  });
}

function configMiddlewares(middlewares) {
  if (middlewares.length > 0) {
    var ccMiddlewares = _ccContext["default"].middlewares;
    ccMiddlewares.length = 0; // 防止热加载重复多次载入middlewares

    middlewares.forEach(function (m) {
      return ccMiddlewares.push(m);
    });
  }
}

function configPlugins(plugins) {
  if (plugins.length > 0) {
    var ccPlugins = _ccContext["default"].plugins;
    ccPlugins.length = 0; // 防止热加载重复多次载入plugins

    (0, _plugin.clearCbs)(); // 清理掉已映射好的插件回调

    var pluginNameMap = {};
    plugins.forEach(function (p) {
      ccPlugins.push(p);

      if (p.install) {
        var pluginInfo = p.install(_plugin.on);
        var e = new Error('plugin.install must return result:{name:string, options?:object}');
        if (!pluginInfo) throw e;
        var pluginName = pluginInfo.name;
        if (!pluginName) throw e;
        if (pluginNameMap[pluginName]) throw new Error("pluginName[" + pluginName + "] duplicate");
        pluginNameMap[pluginName] = 1;
      } else {
        throw new Error('a plugin must export install handler!');
      }
    });
    _ccContext["default"].pluginNameMap = pluginNameMap;
  }
}

var _default = {
  configStoreState: configStoreState,
  configRootReducer: configRootReducer,
  configRootComputed: configRootComputed,
  configRootWatch: configRootWatch,
  configRootLifecycle: configRootLifecycle,
  configMiddlewares: configMiddlewares,
  configPlugins: configPlugins
};
exports["default"] = _default;