"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

exports.__esModule = true;
exports.clearCuRefer = clearCuRefer;
exports["default"] = executeDepFns;

var _util = require("../../support/util");

var _constant = require("../../support/constant");

var _extractStateByKeys2 = _interopRequireDefault(require("../state/extract-state-by-keys"));

var _pickDepFns = _interopRequireDefault(require("../base/pick-dep-fns"));

var _dispatch = _interopRequireDefault(require("../base/dispatch"));

var _makeCuObState = _interopRequireDefault(require("../computed/make-cu-ob-state"));

var _executeAsyncCuInfo = _interopRequireDefault(require("../computed/execute-async-cu-info"));

var _makeCuRefObContainer = require("../computed/make-cu-ref-ob-container");

var _computedMap = _interopRequireDefault(require("../../cc-context/computed-map"));

var _watchMap = _interopRequireDefault(require("../../cc-context/watch-map"));

var _internalVars = require("../../cc-context/internal-vars");

var _wakeyUkeyMap = require("../../cc-context/wakey-ukey-map");

/* eslint-disable camelcase */
var noCommit = function noCommit(tip, asIs) {
  return (0, _util.justWarning)(tip + " call commit or commitCu as it is " + asIs);
}; // 记录某个cuRetKey引用过哪些staticCuRetKeys
// 直接引用或者间接引用过staticCuRetKey都会记录在列表内


var modCuRetKey_referStaticCuRetKeys_ = {};
var refCuRetKey_referStaticCuRetKeys_ = {};

function getCuRetKeyRSListMap(sourceType, module, ccUniqueKey) {
  if (sourceType == _constant.CATE_MODULE) {
    return (0, _util.safeGet)(modCuRetKey_referStaticCuRetKeys_, module);
  } else {
    return (0, _util.safeGet)(refCuRetKey_referStaticCuRetKeys_, ccUniqueKey);
  }
}

function getCuRetKeyRSList(cuRetKey, sourceType, module, ccUniqueKey) {
  var map = getCuRetKeyRSListMap(sourceType, module, ccUniqueKey);
  return (0, _util.safeGetArray)(map, cuRetKey);
}

function clearCuRefer() {
  modCuRetKey_referStaticCuRetKeys_ = {};
  refCuRetKey_referStaticCuRetKeys_ = {};
}

function getCuDep(refCtx, sourceType) {
  return sourceType === _constant.CATE_REF ? refCtx.computedDep : _computedMap["default"]._computedDep;
}

function setStateKeyRetKeysMap(refCtx, sourceType, fnType, stateModule, retKey, keys, isKeysDep) {
  if (isKeysDep === void 0) {
    isKeysDep = true;
  }

  if (keys.length === 0) return;
  var modDep, cuModDep;

  if (sourceType === _constant.CATE_REF) {
    // 由ref发起调用，refCtx是肯定有值的
    var computedDep = refCtx.computedDep;
    var depDesc = fnType === _constant.FN_CU ? computedDep : refCtx.watchDep;
    cuModDep = (0, _util.safeGet)(computedDep, stateModule);
    modDep = (0, _util.safeGet)(depDesc, stateModule);
  } else {
    var cuDep = _computedMap["default"]._computedDep;

    var _depDesc = fnType === _constant.FN_CU ? cuDep : _watchMap["default"]._watchDep;

    cuModDep = (0, _util.safeGet)(cuDep, stateModule);
    modDep = (0, _util.safeGet)(_depDesc, stateModule);
  }

  var stateKey2retKeys = (0, _util.safeGet)(modDep, 'stateKey2retKeys');
  var retKey2stateKeys = (0, _util.safeGet)(modDep, 'retKey2stateKeys');

  var updateRelationship = function updateRelationship(depKeys) {
    var stateKeys = (0, _util.safeGetArray)(retKey2stateKeys, retKey);
    depKeys.forEach(function (sKey) {
      var retKeys = (0, _util.safeGetArray)(stateKey2retKeys, sKey); // 此处判断一下retKeys，谨防用户直接在computed里操作obState, 这里拿到的sKey是一堆原型链上key，如`valueOf`等

      if (Array.isArray(retKeys) && !retKeys.includes(retKey)) retKeys.push(retKey);
      if (!stateKeys.includes(sKey)) stateKeys.push(sKey);
    });
  };

  if (isKeysDep) {
    // keys is depKeys
    updateRelationship(keys);
  } else {
    // keys is retKeys, 将retKeys里各自retKey的stateKeys转移给目标retKey
    keys.forEach(function (sourceRetKey) {
      // 这里取的是cu模块的retKey_stateKeys_
      var retKey2stateKeys = (0, _util.safeGet)(cuModDep, 'retKey2stateKeys');
      var sourceStateKeys = retKey2stateKeys[sourceRetKey] || [];
      updateRelationship(sourceStateKeys);
    });
  }
}

function getRetKeyFnMap(refCtx, sourceType, stateModule) {
  // 始终从_computedDep 取retKey_fn_，来判断commitCu提交的retKey是否合法
  if (sourceType === _constant.CATE_REF) {
    return refCtx.computedRetKeyFns;
  } else {
    var moduleDep = _computedMap["default"]._computedDep[stateModule] || {};
    return moduleDep.retKey2fn || {};
  }
}

function mapRSList(cuRetKey, referCuRetKeys, refCtx, ccUniqueKey, sourceType, stateModule) {
  var cuRetKey_referStaticCuRetKeys_ = getCuRetKeyRSListMap(cuRetKey, sourceType, stateModule, ccUniqueKey);
  var retKey2fn = getRetKeyFnMap(refCtx, sourceType, stateModule);
  var referStaticCuRetKeys = (0, _util.safeGetArray)(cuRetKey_referStaticCuRetKeys_, cuRetKey);
  referCuRetKeys.forEach(function (referCuRetKey) {
    var fnDesc = retKey2fn[referCuRetKey]; // 直接引用

    if (fnDesc.isStatic) {
      referStaticCuRetKeys.push(referCuRetKey);
    } else {
      var tmpRSList = (0, _util.safeGetArray)(cuRetKey_referStaticCuRetKeys_, referCuRetKey); // 把引用的referCuRetKey对应的staticCuRetKey列表记录到当前cuRetKey的staticCuRetKey列表记录上
      // 因为computed函数是严格按需执行的，所以此逻辑能够成立

      tmpRSList.forEach(function (staticCuRetKey) {
        return (0, _util.noDupPush)(referStaticCuRetKeys, staticCuRetKey);
      });
    }
  });
}

var STOP_FN = Symbol('sf'); // fnType: computed watch
// sourceType: module ref
// initialDeltaCommittedState 会在整个过程里收集所有的提交状态

function executeDepFns(ref, stateModule, refModule, oldState, finder, committedState, initialNewState, initialDeltaCommittedState, callInfo, isFirstCall, fnType, sourceType, computedContainer, mergeToDelta) {
  if (ref === void 0) {
    ref = {};
  }

  if (mergeToDelta === void 0) {
    mergeToDelta = true;
  }

  var refCtx = ref.ctx;
  var ccUniqueKey = refCtx ? refCtx.ccUniqueKey : ''; // while循环结束后，收集到的所有的新增或更新state

  var committedStateInWhile = {};
  var nextTickCuInfo = {
    sourceType: sourceType,
    ref: ref,
    module: stateModule,
    fns: [],
    fnAsync: [],
    fnRetKeys: [],
    cuRetContainer: computedContainer
  };
  var whileCount = 0;
  var curStateForComputeFn = committedState;
  var hasDelta = false;

  var _loop = function _loop() {
    whileCount++; // 因为beforeMountFlag为true的情况下，finder里调用的pickDepFns会挑出所有函数，
    // 这里必需保证只有第一次循环的时候取isFirstCall的实际值，否则一定取false，（要不然就陷入无限死循环，每一次都是true，每一次都挑出所有dep函数执行）

    var beforeMountFlag = whileCount === 1 ? isFirstCall : false;

    var _finder = finder(curStateForComputeFn, beforeMountFlag),
        pickedFns = _finder.pickedFns,
        setted = _finder.setted,
        changed = _finder.changed,
        retKey2stateKeys = _finder.retKey2stateKeys;

    nextTickCuInfo.retKey2stateKeys = retKey2stateKeys;
    if (!pickedFns.length) return "break";

    var _makeCommitHandler = (0, _util.makeCommitHandler)(),
        commit = _makeCommitHandler.commit,
        getFnCommittedState = _makeCommitHandler.getFnCommittedState;

    var _makeCommitHandler2 = (0, _util.makeCommitHandler)(),
        commitCu = _makeCommitHandler2.commit,
        getRetKeyCu = _makeCommitHandler2.getFnCommittedState,
        clearCu = _makeCommitHandler2.clear;

    pickedFns.forEach(function (_ref) {
      var retKey = _ref.retKey,
          fn = _ref.fn,
          depKeys = _ref.depKeys,
          isLazy = _ref.isLazy;
      var keyInfo = sourceType + " " + fnType + " retKey[" + retKey + "]";
      var tip = keyInfo + " can't"; // 异步计算的初始值

      var initialVal = '';
      var isInitialValSetted = false;

      var innerDispatch = function innerDispatch(immediate, fnOrFnStr, payload, dispatchOptions) {
        return new Promise(function (resolve) {
          var d = function d() {
            return (0, _dispatch["default"])(fnOrFnStr, payload, dispatchOptions, 0, {
              refModule: stateModule
            });
          };

          if (immediate) resolve(d());else setTimeout(function () {
            return resolve(d());
          }, 0);
        });
      };

      var fnCtx = {
        retKey: retKey,
        callInfo: callInfo,
        isFirstCall: isFirstCall,
        commit: commit,
        commitCu: commitCu,
        setted: setted,
        changed: changed,
        // 在sourceType为module时, 如果非首次计算
        // computedContainer只是一个携带defineProperty的计算结果收集容器，没有收集依赖行为
        cuVal: computedContainer,
        committedState: curStateForComputeFn,
        deltaCommittedState: initialDeltaCommittedState,
        stateModule: stateModule,
        refModule: refModule,
        oldState: oldState,
        refCtx: refCtx,
        dispatch: function dispatch() {
          for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
            args[_key] = arguments[_key];
          }

          return innerDispatch.apply(void 0, [false].concat(args));
        },
        dispatchImmediate: function dispatchImmediate() {
          for (var _len2 = arguments.length, args = new Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
            args[_key2] = arguments[_key2];
          }

          return innerDispatch.apply(void 0, [true].concat(args));
        },
        setInitialVal: function setInitialVal() {
          beforeMountFlag && (0, _util.justWarning)("non async " + keyInfo + " call setInitialVal is unnecessary");
        }
      }; // 循环里的首次计算且是自动收集状态，注入代理对象，收集计算&观察依赖

      var needCollectDep = beforeMountFlag && depKeys === '-'; // 用户通过cuVal读取其他计算结果时，记录cuRetKeys，用于辅助下面计算依赖

      var collectedCuRetKeys = []; // 读取newState时，记录stateKeys，用于辅助下面计算依赖

      var collectedDepKeys = []; // 对于computed，首次计算时会替换为obContainer用于收集依赖
      // !!!对于watch，immediate为true才有机会替换为obContainer收集到依赖

      var referInfo = {
        hasAsyncCuRefer: false
      };

      if (needCollectDep) {
        // 替换cuVal，以便动态的收集到computed&watch函数里读取cuVal时计算相关依赖
        fnCtx.cuVal = (0, _makeCuRefObContainer.getSimpleObContainer)(retKey, sourceType, fnType, stateModule, refCtx, collectedCuRetKeys, referInfo);
      }

      if (fnType === _constant.FN_CU) {
        var isCuFnAsync = (0, _util.isAsyncFn)(fn, stateModule + "/" + retKey);

        if (isLazy || isCuFnAsync) {
          // lazyComputed 和 asyncComputed 不能调用commit commitCu，以隔绝副作用
          var asIs = isLazy ? 'lazy' : 'async computed';

          fnCtx.commit = function () {
            return noCommit(tip, asIs);
          };

          fnCtx.commitCu = fnCtx.commit;
          if (isCuFnAsync) fnCtx.setInitialVal = function (val) {
            initialVal = val;
            isInitialValSetted = true; // 这里阻止异步计算函数的首次执行，交给executeAsyncCuInfo去触发

            if (beforeMountFlag) throw STOP_FN;
          };
        }

        if (isLazy) {
          computedContainer[retKey] = (0, _util.makeCuPackedValue)(isLazy, null, true, fn, initialNewState, oldState, fnCtx);
        } else {
          var newStateArg = initialNewState,
              oldStateArg = oldState; // 首次计算时，new 和 old是同一个对象，方便用于收集depKeys

          if (needCollectDep) {
            oldStateArg = (0, _makeCuObState["default"])(initialNewState, collectedDepKeys);
            newStateArg = oldStateArg;
          } // TODO: fnCtx.connectedState 转为代理对象，用于收集到连接模块的依赖
          // 让示例 https://codesandbox.io/s/ref-watch-read-connected-state-prb4v?file=/src/App.js 正常工作
          // 不同的sourceType，创建的connectedState不一样
          // for module: fnCtx.getComputed, fnCtx.getState，
          // 此处会检查模块加载顺序，然后appendState创建一个隐含的key，然后在目标模块创建一个watch函数
          // for ref: fnCtx.connectedState, fnCtx.connectedComputed
          // 确保 (n,o,f)里的n o总是实例的state


          var computedRet; // 异步函数首次执行时才去调用它，仅为了收集依赖

          if (isCuFnAsync) {
            if (beforeMountFlag) {
              fn(newStateArg, oldStateArg, fnCtx)["catch"](function (err) {
                if (err !== STOP_FN) throw err;
              });
            }
          } else {
            computedRet = fn(newStateArg, oldStateArg, fnCtx);
          }

          if (isCuFnAsync || referInfo.hasAsyncCuRefer) {
            // 首次计算时需要赋初始化值
            if (beforeMountFlag) {
              if (!isInitialValSetted) {
                throw new Error("async " + keyInfo + " forget call setInitialVal");
              }

              computedRet = initialVal;
            } else {
              // 不做任何新的计算，还是赋值原来的结果
              // 新的结果等待 asyncComputedMgr 来计算并触发相关实例重渲染
              computedRet = computedContainer[retKey];
            } // 替换掉setInitialVal，使其失效


            fnCtx.setInitialVal = _util.noop;

            fnCtx.commit = function () {
              return noCommit(tip, 'async computed or it refers async computed ret');
            };

            fnCtx.commitCu = fnCtx.commit; // 安排到nextTickCuInfo里，while结束后单独触发它们挨个按需计算

            nextTickCuInfo.fns.push(function () {
              return fn(newStateArg, oldStateArg, fnCtx);
            });
            nextTickCuInfo.fnAsync.push(isCuFnAsync);
            nextTickCuInfo.fnRetKeys.push(retKey);
          } // 记录计算结果


          computedContainer[retKey] = (0, _util.makeCuPackedValue)(false, computedRet);

          if (needCollectDep) {
            // 在computed函数里读取了newState的stateKey，需要将其记录到当前retKey的依赖列表上
            // 以便能够在相应stateKey值改变时，能够正确命中该computed函数
            setStateKeyRetKeysMap(refCtx, sourceType, _constant.FN_CU, stateModule, retKey, collectedDepKeys); // 在computed里读取cuVal里的其他retKey结果, 要将其他retKey对应的stateKeys写到当前retKey的依赖列表上，
            // 以便能够在相应stateKey值改变时，能够正确命中该computed函数

            setStateKeyRetKeysMap(refCtx, sourceType, _constant.FN_CU, stateModule, retKey, collectedCuRetKeys, false);
            mapRSList(retKey, collectedCuRetKeys, refCtx, ccUniqueKey, sourceType, stateModule);
          }
        }
      } else {
        // watch
        var tmpInitialNewState = initialNewState;
        var tmpOldState = oldState; // 首次触发watch时，才传递ob对象，用于收集依赖

        if (needCollectDep) {
          tmpInitialNewState = (0, _makeCuObState["default"])(initialNewState, collectedDepKeys); // new 和 old是同一个对象，方便用于收集depKeys

          tmpOldState = tmpInitialNewState;
        }

        fn(tmpInitialNewState, tmpOldState, fnCtx); // 首次触发watch时, 才记录依赖

        if (needCollectDep) {
          // 在watch函数里读取了newState的stateKey，需要将其记录到当前watch retKey的依赖列表上
          // 以便能够在相应stateKey值改变时，能够正确命中该watch函数
          setStateKeyRetKeysMap(refCtx, sourceType, _constant.FN_WATCH, stateModule, retKey, collectedDepKeys); // 在watch里读取了cuVal里的retKey结果，要将这些retKey对应的stateKey依赖附加到当前watch retKey的依赖列表上，
          // 以便能够在相应stateKey值改变时，能够正确命中该watch函数

          setStateKeyRetKeysMap(refCtx, sourceType, _constant.FN_WATCH, stateModule, retKey, collectedCuRetKeys, false);
        }
      } // refCompute&refWatch 里获取state、moduleState、connectedState的值收集到的depKeys要记录为ref的静态依赖


      if (needCollectDep && sourceType === _constant.CATE_REF) {
        collectedDepKeys.forEach(function (key) {
          return refCtx.__$$staticWaKeys[(0, _wakeyUkeyMap.makeWaKey)(stateModule, key)] = 1;
        }); // 注：refWatch直接读取了moduleComputed 或者 connectedComputed的值时也收集到了依赖
        // 逻辑在updateDep里判断__$$isBM来确定是不是首次触发
      } // 对于模块计算过程，fn里调用committedCu，computedContainer是moduleComputed结果容器，
      // 对于实例计算过程，fn里调用committedCu来说，computedContainer是refComputed结果容器
      // 每一个retKey返回的committedCu都及时处理掉，因为下面setStateKeyRetKeysMap需要对此时的retKey写依赖


      var committedCuRet = getRetKeyCu();

      if (committedCuRet) {
        var retKey2fn = getRetKeyFnMap(refCtx, sourceType, stateModule);
        (0, _util.okeys)(committedCuRet).forEach(function (cuRetKey) {
          // 模块计算函数里调用commitCu只能修改模块计算retKey
          // 实例计算函数里调用commitCu只能修改实例计算retKey
          var fnDesc = retKey2fn[cuRetKey];
          if (!fnDesc) (0, _util.justWarning)("commitCu:" + tip + " commit [" + cuRetKey + "], it is not defined"); // 由committedCu提交的值，可以统一当作非lazy值set回去，方便取的时候直接取
          else {
            // 检查提交目标只能是静态的cuRetKey
            if (fnDesc.isStatic) {
              var RSList = getCuRetKeyRSList(cuRetKey, sourceType, stateModule, ccUniqueKey);

              if (RSList.includes(cuRetKey)) {
                // 直接或间接引用了这个cuRetKey，就不能去改变它，以避免死循环
                (0, _util.justWarning)("commitCu:" + tip + " change [" + cuRetKey + "], [" + retKey + "] referred [" + cuRetKey + "]");
              } else {
                computedContainer[cuRetKey] = (0, _util.makeCuPackedValue)(false, committedCuRet[cuRetKey]);
              }
            } else {
              (0, _util.justWarning)("commitCu:" + tip + " change [" + cuRetKey + "], it must have zero dep keys");
            }
          }
        });
        clearCu();
      }
    }); // 这里一次性处理所有computed or watch函数提交了然后合并后的state

    curStateForComputeFn = getFnCommittedState();

    if (curStateForComputeFn) {
      // toAssign may be null
      var assignCuState = function assignCuState(toAssign, mergeAssign) {
        if (mergeAssign === void 0) {
          mergeAssign = false;
        }

        // 确保finder函数只针对这一部分新提交的状态去触发computed or watch
        if (mergeAssign) Object.assign(curStateForComputeFn, toAssign);else curStateForComputeFn = toAssign;
        if (!curStateForComputeFn) return;
        Object.assign(committedStateInWhile, curStateForComputeFn);

        if (mergeToDelta) {
          Object.assign(initialNewState, curStateForComputeFn);
          Object.assign(initialDeltaCommittedState, curStateForComputeFn);
        } else {
          // 强行置为null，结束while循环  
          // mergeToDelta为false表示这是来自connectedRefs触发的 cu 或者 wa 函数
          // 此时传入的 initialDeltaCommittedState 是模块state
          // 但是实例里 cu 或 wa 函数只能commit private state
          // 收集到 committedStateInWhile 后，在外面单独触发新的 computedForRef watchForRef过程
          curStateForComputeFn = null;
        }

        hasDelta = true;
      };

      var ensureCommittedState = function ensureCommittedState(fnCommittedState) {
        // !!! 确保实例里调用commit只能提交privState片段，模块里调用commit只能提交moduleState片段
        // !!! 同时确保privState里的key是事先声明过的，而不是动态添加的
        var stateKeys = sourceType === 'ref' ? refCtx.privStateKeys : _internalVars.moduleName2stateKeys[stateModule];

        var _extractStateByKeys = (0, _extractStateByKeys2["default"])(fnCommittedState, stateKeys, true),
            partialState = _extractStateByKeys.partialState,
            ignoredStateKeys = _extractStateByKeys.ignoredStateKeys;

        if (ignoredStateKeys.length) {
          var reason = "they are not " + (sourceType === _constant.CATE_REF ? 'private' : 'module') + " keys, fn is " + sourceType + " " + fnType;
          (0, _util.justWarning)("these state keys[" + ignoredStateKeys.join(',') + "] are invalid, " + reason);
        }

        return partialState; // 返回合法的提交状态
      };

      var partialState = ensureCommittedState(curStateForComputeFn);

      if (partialState) {
        assignCuState(partialState); // watch里提交了新的片段state，再次过一遍computed、watch函数

        if (fnType === _constant.FN_WATCH) {
          // const stateKey2retKeys = getStateKeyRetKeysMap(refCtx, sourceType, stateModule);
          var computedDep = getCuDep(refCtx, sourceType, stateModule);

          var _finder2 = function _finder2(committedState, isBeforeMount) {
            return (0, _pickDepFns["default"])(isBeforeMount, sourceType, _constant.FN_CU, computedDep, stateModule, oldState, committedState, ccUniqueKey);
          }; // 一轮watch函数执行结束，去触发对应的computed计算


          var _executeDepFns = executeDepFns(ref, stateModule, refModule, oldState, _finder2, partialState, initialNewState, initialDeltaCommittedState, callInfo, false, // 再次由watch发起的computed函数查找调用，irFirstCall，一定是false
          _constant.FN_CU, sourceType, computedContainer),
              _hasDelta = _executeDepFns.hasDelta,
              newCommittedState = _executeDepFns.newCommittedState;

          if (_hasDelta) {
            // see https://codesandbox.io/s/complex-cu-watch-chain-s9wzt, 
            // 输入 cc.setState('test', {k1:Date.now()})，确保k4 watch被触发
            var validCommittedState = ensureCommittedState(newCommittedState); // 让validCommittedState合并到curStateForComputeFn里，确保下一轮循环相关watch能被computed里提交的状态触发

            assignCuState(validCommittedState, true);
          }
        }
      }
    }

    if (whileCount > 2) {
      (0, _util.justWarning)('fnCtx.commit may goes endless loop, please check your code');
      (0, _util.justWarning)(callInfo); // 清空，确保不再触发while循环

      curStateForComputeFn = null;
    }
  };

  while (curStateForComputeFn) {
    var _ret = _loop();

    if (_ret === "break") break;
  }

  (0, _executeAsyncCuInfo["default"])(nextTickCuInfo);
  return {
    hasDelta: hasDelta,
    newCommittedState: committedStateInWhile
  };
}