"use strict";

exports.__esModule = true;
exports.clearCachedData = clearCachedData;
exports["default"] = _default;

var _util = require("../../support/util");

/* eslint-disable camelcase */
function getCacheDataContainer() {
  return {
    module: {
      computed: {},
      watch: {}
    },
    ref: {
      computed: {},
      watch: {},
      effect: {}
    }
  };
}

var cacheArea_pickedRetKeys_ = getCacheDataContainer();

function _wrapFn(retKey, retKey2fn, isLazy) {
  var _retKey2fn$retKey = retKey2fn[retKey],
      fn = _retKey2fn$retKey.fn,
      depKeys = _retKey2fn$retKey.depKeys,
      sort = _retKey2fn$retKey.sort;
  return {
    retKey: retKey,
    fn: fn,
    depKeys: depKeys,
    isLazy: isLazy,
    sort: sort
  };
} // asc sort


var sortCb = function sortCb(o1, o2) {
  return o1.sort - o2.sort;
};

function clearCachedData() {
  cacheArea_pickedRetKeys_ = getCacheDataContainer();
} // cate module | ref
// type computed | watch


function _default(isBeforeMount, cate, type, depDesc, stateModule, oldState, committedState, cUkey) {
  var moduleDep = depDesc[stateModule]; // it can be refModuleDep or moduleDep

  var pickedFns = []; // 针对type module， init-module-state时，已对_computedValueOri赋值了默认cuDesc，
  // 所以此时可以安全的直接判断非关系，而不用担心 {}对象存在

  if ((0, _util.isObjectNull)(moduleDep)) return {
    pickedFns: pickedFns,
    setted: [],
    changed: [],
    retKey2stateKeys: {}
  };
  var retKey2fn = moduleDep.retKey2fn,
      retKey2lazy = moduleDep.retKey2lazy,
      stateKey2retKeys = moduleDep.stateKey2retKeys,
      retKey2stateKeys = moduleDep.retKey2stateKeys,
      fnCount = moduleDep.fnCount;
  /** 首次调用 */

  if (isBeforeMount) {
    var retKeys = (0, _util.okeys)(retKey2fn);

    var _setted = (0, _util.okeys)(committedState);

    var _changed = _setted;

    if (type === 'computed') {
      return {
        pickedFns: retKeys.map(function (retKey) {
          return _wrapFn(retKey, retKey2fn, retKey2lazy[retKey]);
        }).sort(sortCb),
        setted: _setted,
        changed: _changed,
        retKey2stateKeys: retKey2stateKeys
      };
    } // for watch


    retKeys.forEach(function (retKey) {
      var _retKey2fn$retKey2 = retKey2fn[retKey],
          fn = _retKey2fn$retKey2.fn,
          immediate = _retKey2fn$retKey2.immediate,
          depKeys = _retKey2fn$retKey2.depKeys,
          sort = _retKey2fn$retKey2.sort;
      if (immediate) pickedFns.push({
        retKey: retKey,
        fn: fn,
        depKeys: depKeys,
        sort: sort
      });
    });
    pickedFns.sort(sortCb);
    return {
      pickedFns: pickedFns,
      setted: _setted,
      changed: _changed,
      retKey2stateKeys: retKey2stateKeys
    };
  } // 这些目标stateKey的值发生了变化


  var _differStateKeys = (0, _util.differStateKeys)(oldState, committedState),
      setted = _differStateKeys.setted,
      changed = _differStateKeys.changed;

  if (setted.length === 0) {
    return {
      pickedFns: pickedFns,
      setted: [],
      changed: [],
      retKey2stateKeys: {}
    };
  } //用setted + changed + module 作为键，缓存对应的pickedFns，这样相同形状的committedState再次进入此函数时，方便快速直接命中pickedFns


  var cacheKey = setted.join(',') + "|" + changed.join(',') + "|" + stateModule; // 要求用户必须在setup里静态的定义完computed & watch，动态的调用computed & watch的回调因为缓存原因不会被触发

  var tmpNode = cacheArea_pickedRetKeys_[cate][type];
  var cachePool = cUkey ? (0, _util.safeGet)(tmpNode, cUkey) : tmpNode;
  var cachedPickedRetKeys = cachePool[cacheKey];

  if (cachedPickedRetKeys) {
    // todo, for 2.5, call checkFnByDepPath with variable depKey_pathDepKeys_
    return {
      pickedFns: cachedPickedRetKeys.map(function (retKey) {
        return _wrapFn(retKey, retKey2fn, retKey2lazy[retKey]);
      }),
      setted: setted,
      changed: changed,
      retKey2stateKeys: retKey2stateKeys
    };
  }

  _pickFn(pickedFns, setted, changed, retKey2fn, stateKey2retKeys, retKey2lazy, fnCount);

  cachePool[cacheKey] = pickedFns.map(function (v) {
    return v.retKey;
  }); // todo, for 2.5, call checkFnByDepPath with variable depKey_pathDepKeys_

  return {
    pickedFns: pickedFns,
    setted: setted,
    changed: changed,
    retKey2stateKeys: retKey2stateKeys
  };
}

function _pickFn(pickedFns, settedStateKeys, changedStateKeys, retKey2fn, stateKey2retKeys, retKey2lazy, fnCount) {
  if (settedStateKeys.length === 0) return; // 把*的函数先全部挑出来, 有key的值发生变化了或者有设值行为

  var starRetKeys = stateKey2retKeys['*'];

  if (starRetKeys) {
    var isKeyValChanged = changedStateKeys.length > 0;
    starRetKeys.forEach(function (retKey) {
      var _retKey2fn$retKey3 = retKey2fn[retKey],
          fn = _retKey2fn$retKey3.fn,
          compare = _retKey2fn$retKey3.compare,
          depKeys = _retKey2fn$retKey3.depKeys,
          sort = _retKey2fn$retKey3.sort;
      var toPush = {
        retKey: retKey,
        fn: fn,
        depKeys: depKeys,
        isLazy: retKey2lazy[retKey],
        sort: sort
      };

      if (compare) {
        if (isKeyValChanged) pickedFns.push(toPush);
        return;
      }

      pickedFns.push(toPush);
    });
  } // 继续遍历settedStateKeys, 挑选出剩余的目标fn（非*相关的）


  if (pickedFns.length < fnCount) {
    (function () {
      var retKey_picked_ = {};
      var len = settedStateKeys.length;

      var _loop = function _loop(i) {
        var stateKey = settedStateKeys[i];
        var retKeys = stateKey2retKeys[stateKey]; //发生变化了的stateKey不一定在依赖列表里

        if (!retKeys) return "continue";
        retKeys.forEach(function (retKey) {
          //没有挑过的方法才挑出来
          if (!retKey_picked_[retKey]) {
            var _retKey2fn$retKey4 = retKey2fn[retKey],
                fn = _retKey2fn$retKey4.fn,
                compare = _retKey2fn$retKey4.compare,
                depKeys = _retKey2fn$retKey4.depKeys,
                sort = _retKey2fn$retKey4.sort;
            var canPick;
            var isValChanged = changedStateKeys.includes(stateKey); // 检测出发生了变化，就一定pick

            if (isValChanged) {
              canPick = true;
            } else {
              // 对于未采用 immutable写法的object是检测不出是否改变的，
              // 因为指向同一个引用，isValChanged一定是false
              // 所以如果compare 为true，则要求用户严格采用immutable写法
              // 为false的话，进入到这里，是已经set的key，canPick一定为true
              canPick = compare ? isValChanged : true;
            }

            if (canPick) {
              retKey_picked_[retKey] = true;
              pickedFns.push({
                retKey: retKey,
                fn: fn,
                depKeys: depKeys,
                isLazy: retKey2lazy[retKey],
                sort: sort
              });
            }
          }
        });
        if (pickedFns.length === fnCount) return "break";
      };

      for (var i = 0; i < len; i++) {
        var _ret = _loop(i);

        if (_ret === "continue") continue;
        if (_ret === "break") break;
      }
    })();
  }

  pickedFns.sort(sortCb);
}