"use strict";

exports.__esModule = true;
exports.exitChain = exitChain;
exports.getAllChainStateMap = getAllChainStateMap;
exports.getChainId = getChainId;
exports.getChainStateList = getChainStateList;
exports.getChainStateMap = getChainStateMap;
exports.isChainExited = isChainExited;
exports.isChainIdLazy = isChainIdLazy;
exports.removeAllChainState = removeAllChainState;
exports.removeChainState = removeChainState;
exports.setAllChainState = setAllChainState;
exports.setAndGetChainStateList = setAndGetChainStateList;
exports.setChainIdLazy = setChainIdLazy;
exports.setChainState = setChainState;

var _util = require("../../support/util");

/* eslint-disable camelcase */
var id = 0;
/** 针对lazy的reducer调用链状态记录缓存map */

var chainId2moduleStateMap = {};
var chainId2isExited = {};
var chainId2isLazy = {};
/** 所有的reducer调用链状态记录缓存map */

var normalChainId_moduleStateMap_ = {};

function getChainId() {
  id++;
  return id;
}

function __setChainState(chainId, targetModule, partialState, targetId_msMap) {
  if (partialState) {
    var moduleStateMap = targetId_msMap[chainId];

    if (!moduleStateMap) {
      moduleStateMap = {};
      targetId_msMap[chainId] = moduleStateMap;
    }

    var state = moduleStateMap[targetModule];

    if (!state) {
      moduleStateMap[targetModule] = partialState;
    } else {
      Object.assign(state, partialState);
    }
  }
}

function setChainState(chainId, targetModule, partialState) {
  __setChainState(chainId, targetModule, partialState, chainId2moduleStateMap);
}

function setAllChainState(chainId, targetModule, partialState) {
  __setChainState(chainId, targetModule, partialState, normalChainId_moduleStateMap_);
}

function setAndGetChainStateList(isC2Result, chainId, targetModule, partialState) {
  if (!isC2Result) setChainState(chainId, targetModule, partialState);
  return getChainStateList(chainId);
}

function getChainStateMap(chainId) {
  return chainId2moduleStateMap[chainId];
}

function getAllChainStateMap(chainId) {
  return normalChainId_moduleStateMap_[chainId];
}

function getChainStateList(chainId) {
  var moduleStateMap = getChainStateMap(chainId);
  return (0, _util.okeys)(moduleStateMap).map(function (m) {
    return {
      module: m,
      state: moduleStateMap[m]
    };
  });
}

function removeChainState(chainId) {
  delete chainId2moduleStateMap[chainId];
}

function removeAllChainState(chainId) {
  delete normalChainId_moduleStateMap_[chainId];
}

function exitChain(chainId) {
  chainId2isExited[chainId] = true;
  removeChainState(chainId);
}

function isChainExited(chainId) {
  return chainId2isExited[chainId] === true;
}

function setChainIdLazy(chainId) {
  chainId2isLazy[chainId] = true;
}

function isChainIdLazy(chainId) {
  return chainId2isLazy[chainId] === true;
}