"use strict";

exports.__esModule = true;
exports["default"] = _default;

var _util = require("../../support/util");

/**
 * 用于传递给 computed 回调收集相关依赖
 * defComputed((newState, oldState)=>{
 *   // 此处的newState oldState即cuObState
 * })
 * @param {{[key:string]:any}} state 
 * @param {string[]} depKeys 
 */
function _default(state, depKeys) {
  return new Proxy(state, {
    get: function get(target, key) {
      /**
       * 第一个isKeyValid判断，是为了防止误使用state算computed value，而触发了其他的key收集
       *   ctx.computed('count', n => {
       *     return n * 2;// 正确写法本应该是 return n.count * 2
       *    })
       *   // 本应该是 n.count * 2, 写为 n * 2 后，触发的key分别为
       *   // valueOf, toString, Symbol(...)
       */
      if ((0, _util.isKeyValid)(target, key) && !depKeys.includes(key)) depKeys.push(key);
      var val = target[key]; // TODO 2_level_key_collect

      return val;
    },
    // set: function (target, key) {
    set: function set() {
      // do nothing，拒绝用户在computed回调里修改state的值
      return true;
    }
  });
}