"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

exports.__esModule = true;
exports.bindEventHandlerToCcContext = bindEventHandlerToCcContext;
exports.findEventHandlersToOff = findEventHandlersToOff;
exports.findEventHandlersToPerform = findEventHandlersToPerform;
exports.getEventItem = getEventItem;
exports.offEventHandlersByCcUniqueKey = offEventHandlersByCcUniqueKey;

var util = _interopRequireWildcard(require("../../support/util"));

var _constant = require("../../support/constant");

var _privConstant = require("../../support/priv-constant");

var _ccContext = _interopRequireDefault(require("../../cc-context"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function _getRequireWildcardCache(nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { "default": obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj["default"] = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

var event2handlers = _ccContext["default"].event2handlers,
    handlerKey2handler = _ccContext["default"].handlerKey2handler,
    ccUKey2handlerKeys = _ccContext["default"].ccUKey2handlerKeys,
    ccUKey2ref = _ccContext["default"].ccUKey2ref;
var makeHandlerKey = util.makeHandlerKey,
    safeGetArray = util.safeGetArray,
    justWarning = util.justWarning;

function _findEventHandlers(event, module, ccClassKey, ccUniqueKey, identity) {
  // 不用默认参数写法了
  // codesandbox lost default value
  var _identity = identity == undefined ? null : identity; // 查找的时候，只负责取，不负责隐式的生成，此次不需要用safeGetArray


  var handlers = event2handlers[event];

  if (handlers) {
    var filteredHandlers = handlers;
    if (ccUniqueKey) filteredHandlers = handlers.filter(function (v) {
      return v.ccUniqueKey === ccUniqueKey;
    });else if (ccClassKey) filteredHandlers = handlers.filter(function (v) {
      return v.ccClassKey === ccClassKey;
    });else if (module) filteredHandlers = handlers.filter(function (v) {
      return v.module === module;
    }); // identity is null means user call emit like emit('eventName')
    // identity is not null means user call emit like emit(['eventName', 'idtName'])

    if (_identity !== undefined) {
      filteredHandlers = filteredHandlers.filter(function (v) {
        return v.identity === _identity;
      });
    }

    return filteredHandlers;
  }

  return [];
}

function _deleteEventHandlers(handlers) {
  var toDeleteHandlerKeyMap = {};
  var toDeleteCcUniqueKeyMap = {};
  var toDeleteEventNames = [];
  handlers.forEach(function (item) {
    var handlerKey = item.handlerKey,
        ccUniqueKey = item.ccUniqueKey,
        event = item.event;
    delete handlerKey2handler[handlerKey]; // delete mapping of handlerKey2handler;

    toDeleteHandlerKeyMap[handlerKey] = 1;
    toDeleteCcUniqueKeyMap[ccUniqueKey] = 1;
    if (!toDeleteEventNames.includes(event)) toDeleteEventNames.push(event);
  });
  toDeleteEventNames.forEach(function (event) {
    var eHandlers = event2handlers[event];

    if (eHandlers) {
      eHandlers.forEach(function (h, idx) {
        var ccUniqueKey = h.ccUniqueKey;

        if (toDeleteCcUniqueKeyMap[ccUniqueKey] === 1) {
          eHandlers[idx] = null;
          delete ccUKey2handlerKeys[ccUniqueKey]; // delete mapping of ccUKey2handlerKeys;
        }
      });
      event2handlers[event] = eHandlers.filter(function (v) {
        return v !== null;
      }); // delete eHandlers null element
    }
  });
}

function bindEventHandlerToCcContext(module, ccClassKey, ccUniqueKey, event, identity, handler) {
  var handlers = safeGetArray(event2handlers, event);

  if (!util.isFn(handler)) {
    return justWarning("event " + event + "'s handler " + _privConstant.INAF + "!");
  }

  var handlerKey = makeHandlerKey(ccUniqueKey, event, identity);
  var handlerKeys = safeGetArray(ccUKey2handlerKeys, ccUniqueKey);
  var targetHandlerIndex = handlers.findIndex(function (v) {
    return v.handlerKey === handlerKey;
  }); // user call ctx.on for a same event in a same instance more than once

  var handlerItem = {
    event: event,
    module: module,
    ccClassKey: ccClassKey,
    ccUniqueKey: ccUniqueKey,
    identity: identity,
    handlerKey: handlerKey,
    fn: handler
  };

  if (targetHandlerIndex > -1) {
    // will alway use the latest handler
    handlers[targetHandlerIndex] = handlerItem;
  } else {
    handlers.push(handlerItem);
    handlerKeys.push(handlerKey);
  }

  handlerKey2handler[handlerKey] = handlerItem;
}

function findEventHandlersToPerform(event) {
  for (var _len = arguments.length, args = new Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
    args[_key - 1] = arguments[_key];
  }

  var _event,
      _identity = null,
      _module,
      _ccClassKey,
      _ccUniqueKey;

  var canPerform = null;

  if (typeof event === 'string') {
    _event = event;
  } else {
    _event = event.name;
    _identity = event.identity;
    _module = event.module;
    _ccClassKey = event.ccClassKey;
    _ccUniqueKey = event.ccUniqueKey;
    canPerform = event.canPerform;
  }

  var handlers = _findEventHandlers(_event, _module, _ccClassKey, _ccUniqueKey, _identity);

  handlers.forEach(function (_ref) {
    var ccUniqueKey = _ref.ccUniqueKey,
        handlerKey = _ref.handlerKey;
    var ref = ccUKey2ref[ccUniqueKey];

    if (ref && handlerKey) {
      // confirm the instance is mounted and handler is not been offed
      if (ref.__$$ms !== _constant.MOUNTED) return;
      var handler = handlerKey2handler[handlerKey];

      if (handler) {
        if (canPerform && !canPerform(ref)) {
          return;
        }

        handler.fn.apply(handler, args);
      }
    }
  });
}

function findEventHandlersToOff(event, _ref2) {
  var module = _ref2.module,
      ccClassKey = _ref2.ccClassKey,
      ccUniqueKey = _ref2.ccUniqueKey,
      identity = _ref2.identity;

  var handlers = _findEventHandlers(event, module, ccClassKey, ccUniqueKey, identity);

  _deleteEventHandlers(handlers);
}

function offEventHandlersByCcUniqueKey(ccUniqueKey) {
  var handlerKeys = ccUKey2handlerKeys[ccUniqueKey];

  if (handlerKeys) {
    var toDeleteHandlers = [];
    handlerKeys.forEach(function (k) {
      return toDeleteHandlers.push(handlerKey2handler[k]);
    });

    _deleteEventHandlers(toDeleteHandlers);
  }
}

function getEventItem(event) {
  var outputEv;

  if (event && typeof event === 'object') {
    var _event;

    if (Array.isArray(event)) {
      var name = event[0],
          identity = event[1];
      _event = {
        name: name,
        identity: identity
      };
    } else {
      _event = Object.assign({}, event);
    }

    if (!_event.identity) _event.identity = null; //否则就允许用户传如自己定义的module, ccClassKey

    outputEv = _event;
  } else {
    outputEv = {
      name: event,
      identity: null
    };
  }

  return outputEv;
}