"use strict";

exports.__esModule = true;
exports.checkConnectSpec = checkConnectSpec;
exports.checkKeys = checkKeys;
exports.checkModuleName = checkModuleName;
exports.checkModuleNameAndState = checkModuleNameAndState;
exports.checkModuleNameBasically = checkModuleNameBasically;
exports.checkRenderKeyClasses = checkRenderKeyClasses;
exports.checkStoredKeys = checkStoredKeys;

var util = _interopRequireWildcard(require("../../support/util"));

var _privConstant = require("../../support/priv-constant");

var _constant = require("../../support/constant");

var _internalVars = require("../../cc-context/internal-vars");

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function _getRequireWildcardCache(nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { "default": obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj["default"] = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

var isModuleNameCcLike = util.isModuleNameCcLike,
    isModuleNameValid = util.isModuleNameValid,
    vbi = util.verboseInfo,
    makeError = util.makeError,
    okeys = util.okeys;
/** 检查模块名，名字合法，就算检查通过 */

function checkModuleNameBasically(moduleName, innerParams) {
  if (innerParams === void 0) {
    innerParams = {};
  }

  if (!isModuleNameValid(moduleName)) {
    throw new Error("module[" + moduleName + "] writing is invalid!");
  } // moduleName will be named starting with $cc while calling createModule


  if (innerParams[_privConstant.ALCC_KEY] === 1) return;

  if (isModuleNameCcLike(moduleName)) {
    throw new Error("'$$cc' is a built-in module name");
  }
}
/**
 * 检查模块名, moduleMustNotExisted 默认为true，
 * true表示【module名字合法】且【对应的moduleState不存在】，才算检查通过  
 * false表示【module名字合法】且【对应的moduleState存在】，才算检查通过
 * @param {string} moduleName 
 * @param {boolean} [moduleMustNotExisted=true] - true 要求模块应该不存在 ,false 要求模块状态应该已存在
 */


function checkModuleName(moduleName, moduleMustNotExisted, vbiMsg, innerParams) {
  if (moduleMustNotExisted === void 0) {
    moduleMustNotExisted = true;
  }

  if (vbiMsg === void 0) {
    vbiMsg = '';
  }

  var _vbiMsg = vbiMsg || "module[" + moduleName + "]";

  checkModuleNameBasically(moduleName, innerParams);

  if (moduleName !== _constant.MODULE_GLOBAL) {
    if (moduleMustNotExisted) {
      if (util.isObjectNotNull(_internalVars._state[moduleName])) {
        // 但是却存在了
        throw makeError(_constant.ERR.CC_MODULE_NAME_DUPLICATE, vbi(_vbiMsg));
      }
    } else {
      if (!_internalVars._state[moduleName]) {
        // 实际上却不存在
        throw makeError(_constant.ERR.CC_MODULE_NOT_FOUND, vbi(_vbiMsg));
      }
    }
  }
}

function checkModuleNameAndState(moduleName, moduleState, moduleMustNotExisted, innerParams) {
  checkModuleName(moduleName, moduleMustNotExisted, '', innerParams);

  if (!util.isPJO(moduleState)) {
    throw new Error("module[" + moduleName + "]'s state " + _privConstant.INAJ);
  }
}

function checkStoredKeys(belongModule, storedKeys) {
  if (storedKeys === '*') {
    return;
  }

  if (Array.isArray(storedKeys)) {
    checkKeys(belongModule, storedKeys, false, 'storedKeys invalid ');
    return;
  }

  throw new Error("storedKeys type err, " + _privConstant.STR_ARR_OR_STAR);
}

function checkKeys(module, keys, keyShouldBeModuleStateKey, extraInfo) {
  if (keyShouldBeModuleStateKey === void 0) {
    keyShouldBeModuleStateKey = true;
  }

  if (extraInfo === void 0) {
    extraInfo = '';
  }

  var keyword = keyShouldBeModuleStateKey ? '' : 'not ';

  var keyTip = function keyTip(name, keyword) {
    return extraInfo + "key[" + name + "] must " + keyword + "be a module state key";
  };

  var moduleStateKeys = _internalVars.moduleName2stateKeys[module] || [];
  keys.forEach(function (sKey) {
    var keyInModuleState = moduleStateKeys.includes(sKey);

    var throwErr = function throwErr() {
      throw new Error(keyTip(sKey, keyword));
    };

    if (keyShouldBeModuleStateKey) {
      !keyInModuleState && throwErr();
    } else {
      keyInModuleState && throwErr();
    }
  });
}

function checkConnectSpec(connectSpec) {
  var invalidConnect = "param connect is invalid,";

  var invalidConnectItem = function invalidConnectItem(m) {
    return invalidConnect + " module[" + m + "]'s value " + _privConstant.STR_ARR_OR_STAR;
  };

  okeys(connectSpec).forEach(function (m) {
    checkModuleName(m, false);
    var val = connectSpec[m];

    if (typeof val === 'string') {
      if (val !== '*' && val !== '-') throw new Error(invalidConnectItem(m));
    } else if (!Array.isArray(val)) {
      throw new Error(invalidConnectItem(m));
    } else {
      checkKeys(m, val, true, "connect module[" + m + "] invalid,");
    }
  });
}

function checkRenderKeyClasses(regRenderKeyClasses) {
  if (!Array.isArray(regRenderKeyClasses) && regRenderKeyClasses !== '*') {
    throw new Error("renderKeyClasses type err, it " + _privConstant.STR_ARR_OR_STAR);
  }
}