"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

exports.__esModule = true;
exports.clearCbs = clearCbs;
exports.on = on;
exports.send = send;

var _constant = require("../../support/constant");

var _runtimeHandler = _interopRequireDefault(require("../../cc-context/runtime-handler"));

var sigs = [_constant.SIG_FN_START, _constant.SIG_FN_END, _constant.SIG_FN_QUIT, _constant.SIG_FN_ERR, _constant.SIG_MODULE_CONFIGURED, _constant.SIG_STATE_CHANGED, _constant.SIG_ASYNC_COMPUTED_START, _constant.SIG_ASYNC_COMPUTED_END, _constant.SIG_ASYNC_COMPUTED_ERR, _constant.SIG_ASYNC_COMPUTED_BATCH_START, _constant.SIG_ASYNC_COMPUTED_BATCH_END];
var sig2cbs = {};
sigs.forEach(function (sig) {
  return sig2cbs[sig] = [];
});

function _pushSigCb(sigMap, sigOrSigs, cb) {
  function pushCb(sig, cb) {
    var cbs = sigMap[sig];

    if (cb) {
      cbs.push(cb);
    } else {
      console.warn("invalid sig[" + sig + "]");
    }
  }

  if (Array.isArray(sigOrSigs)) {
    sigOrSigs.forEach(function (sig) {
      pushCb(sig, cb);
    });
  } else {
    pushCb(sigOrSigs, cb);
  }
}

function clearCbs() {
  sigs.forEach(function (sig) {
    return sig2cbs[sig].length = 0;
  });
}

function send(sig, payload) {
  var cbs = sig2cbs[sig];
  cbs.forEach(function (cb) {
    try {
      cb({
        sig: sig,
        payload: payload
      });
    } catch (err) {
      // plugin error should not abort dispatch process
      // for letting plugin error be isolate, I have to put try catch block in forEach loop
      _runtimeHandler["default"].tryHandleError(err, true);
    }
  });
}

function on(sigOrSigs, cb) {
  _pushSigCb(sig2cbs, sigOrSigs, cb);
}