"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

exports.__esModule = true;
exports.clearCount = clearCount;
exports.decCcKeyInsCount = decCcKeyInsCount;
exports["default"] = _default;
exports.getCcKeyInsCount = getCcKeyInsCount;
exports.incCcKeyInsCount = incCcKeyInsCount;

var _ccContext = _interopRequireDefault(require("../../cc-context"));

var _constant = require("../../support/constant");

var util = _interopRequireWildcard(require("../../support/util"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function _getRequireWildcardCache(nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { "default": obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj["default"] = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

/* eslint-disable camelcase */
var justWarning = util.justWarning,
    me = util.makeError,
    vbi = util.verboseInfo,
    ss = util.styleStr,
    cl = util.color,
    logNormal = util.logNormal;
var runtimeVar = _ccContext["default"].runtimeVar,
    ccUKey2ref = _ccContext["default"].ccUKey2ref;
var ccUKey2insCount = {};

function setCcInstanceRef(ccUniqueKey, ref, delayMs) {
  var setRef = function setRef() {
    ccUKey2ref[ccUniqueKey] = ref;
  };

  if (_ccContext["default"].isHotReloadMode()) incCcKeyInsCount(ccUniqueKey);

  if (delayMs) {
    setTimeout(setRef, delayMs);
  } else {
    setRef();
  }
}

function incCcKeyInsCount(ccUniqueKey) {
  util.safeAdd(ccUKey2insCount, ccUniqueKey, 1);
}

function decCcKeyInsCount(ccUniqueKey) {
  util.safeMinus(ccUKey2insCount, ccUniqueKey, 1);
}

function getCcKeyInsCount(ccUniqueKey) {
  return ccUKey2insCount[ccUniqueKey] || 0;
}

function clearCount() {
  ccUKey2insCount = {};
}

function _default(ref) {
  var _ref$ctx = ref.ctx,
      ccClassKey = _ref$ctx.ccClassKey,
      ccKey = _ref$ctx.ccKey,
      ccUniqueKey = _ref$ctx.ccUniqueKey;

  if (runtimeVar.isDebug) {
    logNormal(ss("register ccKey " + ccUniqueKey + " to CC_CONTEXT"), cl());
  }

  var isHot = _ccContext["default"].isHotReloadMode();

  if (ccUKey2ref[ccUniqueKey]) {
    var dupErr = function dupErr() {
      throw me(_constant.ERR.CC_CLASS_INSTANCE_KEY_DUPLICATE, vbi("ccClass:" + ccClassKey + ",ccKey:" + ccKey));
    };

    if (isHot) {
      // get existed ins count
      var insCount = getCcKeyInsCount(ccUniqueKey);

      if (insCount > 1) {
        // now cc can make sure the ccKey duplicate
        dupErr();
      } // just warning


      justWarning("\n        found ccKey[" + ccKey + "] duplicated in hot reload mode, please make sure your ccKey is unique manually,\n        " + vbi("ccClassKey:" + ccClassKey + " ccKey:" + ccKey + " ccUniqueKey:" + ccUniqueKey) + "\n      "); // in webpack hot reload mode, cc works not very well,
      // cc can't set ref immediately, because the ccInstance of ccKey will ummount right now in unmount func, 
      // cc call unsetCcInstanceRef will lost the right ref in CC_CONTEXT.refs
      // so cc set ref later

      setCcInstanceRef(ccUniqueKey, ref, 600);
    } else {
      dupErr();
    }
  } else {
    setCcInstanceRef(ccUniqueKey, ref);
  }
}