"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

exports.__esModule = true;
exports["default"] = startChangeRefState;

var util = _interopRequireWildcard(require("../../support/util"));

var cst = _interopRequireWildcard(require("../../support/constant"));

var _privConstant = require("../../support/priv-constant");

var _runLater = _interopRequireDefault(require("../base/run-later"));

var _ccContext = _interopRequireDefault(require("../../cc-context"));

var _extractStateByKeys4 = _interopRequireDefault(require("../state/extract-state-by-keys"));

var _watchKeyForRef2 = _interopRequireDefault(require("../watch/watch-key-for-ref"));

var _computeValueForRef2 = _interopRequireDefault(require("../computed/compute-value-for-ref"));

var _findUpdateRefs2 = _interopRequireDefault(require("../ref/find-update-refs"));

var _plugin = require("../plugin");

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function _getRequireWildcardCache(nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { "default": obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj["default"] = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

/* eslint-disable camelcase */

/** @typedef {import('../../types').ICtxBase} ICtxBase */
var isPJO = util.isPJO,
    justWarning = util.justWarning,
    isObjectNull = util.isObjectNull,
    computeFeature = util.computeFeature,
    okeys = util.okeys;
var FOR_CUR_MOD = cst.FOR_CUR_MOD,
    FOR_ANOTHER_MOD = cst.FOR_ANOTHER_MOD,
    FORCE_UPDATE = cst.FORCE_UPDATE,
    SET_STATE = cst.SET_STATE,
    SIG_STATE_CHANGED = cst.SIG_STATE_CHANGED,
    RENDER_NO_OP = cst.RENDER_NO_OP,
    RENDER_BY_KEY = cst.RENDER_BY_KEY,
    RENDER_BY_STATE = cst.RENDER_BY_STATE,
    UNMOUNTED = cst.UNMOUNTED,
    NOT_MOUNT = cst.NOT_MOUNT;
var _ccContext$store = _ccContext["default"].store,
    storeSetState = _ccContext$store.setState,
    getPrevState = _ccContext$store.getPrevState,
    saveSharedState = _ccContext$store.saveSharedState,
    middlewares = _ccContext["default"].middlewares,
    ccClassKey2Context = _ccContext["default"].ccClassKey2Context,
    refStore = _ccContext["default"].refStore,
    getModuleStateKeys = _ccContext["default"].getModuleStateKeys,
    runtimeVar = _ccContext["default"].runtimeVar; // 触发修改状态的实例所属模块和目标模块不一致的时候，stateFor是 FOR_ANOTHER_MOD

function getStateFor(targetModule, refModule) {
  return targetModule === refModule ? FOR_CUR_MOD : FOR_ANOTHER_MOD;
}

function callMiddlewares(skipMiddleware, passToMiddleware, cb) {
  if (skipMiddleware !== true) {
    var len = middlewares.length;

    if (len > 0) {
      var index = 0;

      var next = function next() {
        if (index === len) {
          // all middlewares been executed
          cb();
        } else {
          var middlewareFn = middlewares[index];
          index++;
          if (util.isFn(middlewareFn)) middlewareFn(passToMiddleware, next);else {
            justWarning("found one middleware " + _privConstant.INAF);
            next();
          }
        }
      };

      next();
    } else {
      cb();
    }
  } else {
    cb();
  }
} // 调用者优先取 alwaysRenderCaller，再去force参数


function getCallerForce(force) {
  return runtimeVar.alwaysRenderCaller || force;
}
/**
 * 修改状态入口函数
 */


function changeRefState(inputState, _temp, callerRef) {
  var _ref = _temp === void 0 ? {} : _temp,
      module = _ref.module,
      _ref$skipMiddleware = _ref.skipMiddleware,
      skipMiddleware = _ref$skipMiddleware === void 0 ? false : _ref$skipMiddleware,
      payload = _ref.payload,
      stateChangedCb = _ref.stateChangedCb,
      _ref$force = _ref.force,
      force = _ref$force === void 0 ? false : _ref$force,
      stateSnapshot = _ref.stateSnapshot,
      _ref$keys = _ref.keys,
      keys = _ref$keys === void 0 ? [] : _ref$keys,
      _ref$keyPath = _ref.keyPath,
      keyPath = _ref$keyPath === void 0 ? '' : _ref$keyPath,
      reactCallback = _ref.reactCallback,
      type = _ref.type,
      _ref$calledBy = _ref.calledBy,
      calledBy = _ref$calledBy === void 0 ? SET_STATE : _ref$calledBy,
      _ref$fnName = _ref.fnName,
      fnName = _ref$fnName === void 0 ? '' : _ref$fnName,
      renderKey = _ref.renderKey,
      _ref$delay = _ref.delay,
      delay = _ref$delay === void 0 ? -1 : _ref$delay;

  if (!inputState) return;

  if (!isPJO(inputState)) {
    return;
  }

  var state = inputState;
  var targetRenderKey = util.extractRenderKey(renderKey);
  var targetDelay = renderKey && renderKey.delay ? renderKey.delay : delay;
  var _callerRef$ctx = callerRef.ctx,
      refModule = _callerRef$ctx.module,
      ccUniqueKey = _callerRef$ctx.ccUniqueKey,
      ccKey = _callerRef$ctx.ccKey;
  var stateFor = getStateFor(module, refModule);
  var callInfo = {
    calledBy: calledBy,
    payload: payload,
    renderKey: targetRenderKey,
    force: force,
    ccKey: ccKey,
    module: module,
    fnName: fnName,
    keys: keys,
    keyPath: keyPath
  }; // 在triggerReactSetState之前把状态存储到store，
  // 防止属于同一个模块的父组件套子组件渲染时，父组件修改了state，子组件初次挂载是不能第一时间拿到state
  // const passedRef = stateFor === FOR_CUR_MOD ? targetRef : null;
  // 标记noSave为true，延迟到后面可能存在的中间件执行结束后才save

  var _syncCommittedStateTo = syncCommittedStateToStore(module, inputState, {
    ref: callerRef,
    callInfo: callInfo,
    noSave: true,
    force: force
  }),
      sharedState = _syncCommittedStateTo.partialState,
      hasDelta = _syncCommittedStateTo.hasDelta,
      hasPrivState = _syncCommittedStateTo.hasPrivState; // 有 computed 提交了新的 state


  if (hasDelta) {
    state = Object.assign({}, inputState, sharedState);
  } // 不包含私有状态，仅包含模块状态，交给 belongRefs 那里去触发渲染，这样可以让已失去依赖的当前实例减少一次渲染
  // 因为 belongRefs 那里是根据有无依赖来确定要不要渲染，这样的话如果失去了依赖不把它查出来就不触发它渲染了


  var ignoreRender = !hasPrivState && !!sharedState; // source ref will receive the whole committed state 

  triggerReactSetState(callerRef, callInfo, targetRenderKey, calledBy, state, stateFor, ignoreRender, reactCallback, getCallerForce(force), function (renderType, committedState, refUpdater) {
    // committedState means final committedState
    var passToMiddleware = {
      calledBy: calledBy,
      type: type,
      payload: payload,
      renderKey: targetRenderKey,
      delay: targetDelay,
      ccKey: ccKey,
      ccUniqueKey: ccUniqueKey,
      committedState: committedState,
      refModule: refModule,
      module: module,
      fnName: fnName,
      sharedState: sharedState || {} // 给一个空壳对象，防止用户直接用的时候报错null

    };
    var modStateCalled = false; // 修改或新增状态值
    // 修改并不会再次触发compute&watch过程，请明确你要修改的目的

    passToMiddleware.modState = function (key, val) {
      modStateCalled = true;
      passToMiddleware.committedState[key] = val;
      passToMiddleware.sharedState[key] = val;
    };

    callMiddlewares(skipMiddleware, passToMiddleware, function () {
      // 到这里才触发调用 saveSharedState 存储模块状态和 并用 refUpdater 更新调用实例，注这两者前后顺序不能调换
      // 因为 callerRef 里的 beforeRender 步骤会把最新的模块状态合进来
      // 允许在中间件过程中使用「modState」修改某些key的值，会影响到实例的更新结果，但不会再触发computed&watch
      // 所以调用此接口请明确知道上面的后果，可能导致以外的bug
      // 注不要直接修改 sharedState 或 committedState，如非要修改，应该是对两个对象一起修改某个key才是正确的
      var midSharedState = passToMiddleware.sharedState; // 如 finalSharedState 为空，表示提交的状态和模块状态没有发生变化

      var finalSharedState = saveSharedState(module, midSharedState, modStateCalled, force); // TODO: 查看其它模块的cu函数里读取了当前模块的state或computed作为输入产生了的新的计算结果
      // 然后做相应的关联更新 {'$$global/key1': {foo: ['cuKey1', 'cuKey2'] } }
      // code here
      // 执行完毕所有的中间件，才更新触发调用的源头实例

      refUpdater && refUpdater();

      if (renderType === RENDER_NO_OP && !finalSharedState) {
        if (ignoreRender) {
          // 此时 refUpdater 为 null, 主动为 caller 执行一次 triggerReactSetState，
          // 以便让 triggerReactSetState 内部有机会触发 reactCallback
          triggerReactSetState(callerRef, callInfo, [], SET_STATE, midSharedState, stateFor, true, reactCallback, getCallerForce(force));
        }
      } else {
        (0, _plugin.send)(SIG_STATE_CHANGED, {
          calledBy: calledBy,
          type: type,
          committedState: committedState,
          sharedState: finalSharedState || {},
          payload: payload,
          module: module,
          ccUniqueKey: ccUniqueKey,
          renderKey: targetRenderKey,
          stateSnapshot: stateSnapshot
        });
      } // 无论是否真的有状态改变，此回调都会被触发


      if (stateChangedCb) stateChangedCb(); // 当前上下文的ignoreRender 为true时， 等效于入参 allowOriInsRender 为true，允许查询出oriIns后触发它渲染

      if (finalSharedState) triggerBroadcastState(stateFor, callInfo, callerRef, finalSharedState, ignoreRender, module, reactCallback, targetRenderKey, targetDelay, force);
    });
  });
}

function triggerReactSetState(callerRef, callInfo, renderKeys, calledBy, state, stateFor, ignoreRender, reactCallback, force, next) {
  var refCtx = callerRef.ctx;
  var refState = refCtx.unProxyState;

  var nextNoop = function nextNoop() {
    next && next(RENDER_NO_OP, state); // fix issue: https://github.com/concentjs/concent/issues/70
    // 没有任何依赖的组件，定义了cb，也让其有机会执行

    if (reactCallback) {
      var __$$mstate = refCtx.__$$mstate;
      var newState = Object.assign({}, refState, __$$mstate);
      reactCallback(newState);
    }
  };

  if (ignoreRender) {
    return nextNoop();
  }

  if (callerRef.__$$ms === UNMOUNTED // 已卸载
  || stateFor !== FOR_CUR_MOD // 确保 forceUpdate 能够刷新cc实例，因为state可能是{}，此时用户调用forceUpdate也要触发render
  || calledBy !== FORCE_UPDATE && isObjectNull(state)) {
    return nextNoop();
  }

  var stateModule = refCtx.module,
      storedKeys = refCtx.storedKeys,
      ccUniqueKey = refCtx.ccUniqueKey;
  var renderType = RENDER_BY_STATE;

  if (renderKeys.length) {
    // if user specify renderKeys
    renderType = RENDER_BY_KEY;

    if (renderKeys.includes(refCtx.renderKey)) {
      // current instance can been rendered only if ctx.renderKey included in renderKeys
      return nextNoop();
    }
  }

  if (storedKeys.length > 0) {
    var _extractStateByKeys = (0, _extractStateByKeys4["default"])(state, storedKeys),
        partialState = _extractStateByKeys.partialState,
        isStateEmpty = _extractStateByKeys.isStateEmpty;

    if (!isStateEmpty) {
      if (refCtx.persistStoredKeys === true) {
        var _extractStateByKeys2 = (0, _extractStateByKeys4["default"])(refState, storedKeys),
            entireStoredState = _extractStateByKeys2.partialState;

        var currentStoredState = Object.assign({}, entireStoredState, partialState);

        if (_ccContext["default"].localStorage) {
          _ccContext["default"].localStorage.setItem("CCSS_" + ccUniqueKey, JSON.stringify(currentStoredState));
        }
      }

      refStore.setState(ccUniqueKey, partialState);
    }
  }

  var deltaCommittedState = Object.assign({}, state);
  (0, _computeValueForRef2["default"])(callerRef, stateModule, refState, deltaCommittedState, callInfo);
  (0, _watchKeyForRef2["default"])(callerRef, stateModule, refState, deltaCommittedState, callInfo);

  var ccSetState = function ccSetState() {
    // 使用 unProxyState ，避免触发get
    var mayChangedState;
    if (force === true) mayChangedState = deltaCommittedState;else mayChangedState = util.extractChangedState(refCtx.unProxyState, deltaCommittedState);

    if (mayChangedState) {
      // 记录 stateKeys，方便 triggerRefEffect 之用
      refCtx.__$$settedList.push({
        module: stateModule,
        keys: okeys(mayChangedState)
      });

      var upCb = function upCb() {
        return refCtx.__$$ccSetState(mayChangedState, reactCallback);
      };

      if (callerRef.__$$ms === NOT_MOUNT) {
        refCtx.__$$queuedUpdaters.push(upCb);
      } else {
        upCb();
      }
    }
  };

  if (next) {
    next(renderType, deltaCommittedState, ccSetState);
  } else {
    ccSetState();
  }
}

function syncCommittedStateToStore(moduleName, committedState, options) {
  var stateKeys = getModuleStateKeys(moduleName); // extract shared state

  var _extractStateByKeys3 = (0, _extractStateByKeys4["default"])(committedState, stateKeys, true),
      partialState = _extractStateByKeys3.partialState,
      hasPrivState = _extractStateByKeys3.missKeyInState; // save state to store


  if (partialState) {
    var _storeSetState = storeSetState(moduleName, partialState, options),
        hasDelta = _storeSetState.hasDelta,
        deltaCommittedState = _storeSetState.deltaCommittedState;

    return {
      partialState: deltaCommittedState,
      hasDelta: hasDelta,
      hasPrivState: hasPrivState
    };
  }

  return {
    partialState: partialState,
    hasDelta: false,
    hasPrivState: hasPrivState
  };
}

function triggerBroadcastState(stateFor, callInfo, targetRef, sharedState, allowOriInsRender, moduleName, reactCallback, renderKeys, delay, force) {
  var passAllowOri = allowOriInsRender;

  if (delay > 0) {
    if (passAllowOri) {
      // 优先将当前实例渲染了
      triggerReactSetState(targetRef, callInfo, [], SET_STATE, sharedState, stateFor, false, reactCallback, getCallerForce(force));
    }

    passAllowOri = false; // 置为false，后面的runLater里不会再次触发当前实例渲染
  }

  var startBroadcastState = function startBroadcastState() {
    broadcastState(callInfo, targetRef, sharedState, passAllowOri, moduleName, reactCallback, renderKeys, force);
  };

  if (delay > 0) {
    var feature = computeFeature(targetRef.ctx.ccUniqueKey, sharedState);
    (0, _runLater["default"])(startBroadcastState, feature, delay);
  } else {
    startBroadcastState();
  }
}

function broadcastState(callInfo, callerRef, partialSharedState, allowOriInsRender, moduleName, reactCallback, renderKeys, force) {
  if (!partialSharedState) {
    // null
    return;
  }

  var ccUKey2ref = _ccContext["default"].ccUKey2ref;
  /** @type ICtxBase */

  var _callerRef$ctx2 = callerRef.ctx,
      currentCcUKey = _callerRef$ctx2.ccUniqueKey,
      ccClassKey = _callerRef$ctx2.ccClassKey;
  var renderKeyClasses = ccClassKey2Context[ccClassKey].renderKeyClasses;

  var _findUpdateRefs = (0, _findUpdateRefs2["default"])(moduleName, partialSharedState, renderKeys, renderKeyClasses),
      sharedStateKeys = _findUpdateRefs.sharedStateKeys,
      _findUpdateRefs$resul = _findUpdateRefs.result,
      belongRefKeys = _findUpdateRefs$resul.belong,
      connectRefKeys = _findUpdateRefs$resul.connect; // 引入更新队列，放入下一轮事件循环更新，避免一下警告
  // Warning: Cannot update a component (`XXX`) while rendering a different component (`YYY`)


  var updaterQueue = [];
  var oriCalledBy = callInfo.calledBy;

  var queuedOrUpdate = function queuedOrUpdate(refKey, updater) {
    // 触发者调用则必须立即触发更新，避免input框录入字符时出现无法输入中文的问题
    if (refKey === currentCcUKey) {
      updater();
    } else {
      updaterQueue.push(updater);
    }
  };

  var renderedInBelong = {};
  belongRefKeys.forEach(function (refKey) {
    var ref = ccUKey2ref[refKey];
    if (!ref) return;
    var rcb = null; // 这里的calledBy直接用'broadcastState'，仅供concent内部运行时用

    var calledBy = 'broadcastState';

    if (refKey === currentCcUKey) {
      if (!allowOriInsRender) return;
      rcb = reactCallback;
      calledBy = oriCalledBy;
    }

    queuedOrUpdate(refKey, function () {
      triggerReactSetState(ref, callInfo, [], calledBy, partialSharedState, FOR_CUR_MOD, false, rcb, force);
    });
    renderedInBelong[refKey] = 1;
  });
  var prevModuleState = getPrevState(moduleName);
  connectRefKeys.forEach(function (refKey) {
    // 对于即属于又连接的实例，避免一次重复的渲染
    if (renderedInBelong[refKey]) {
      return;
    }

    var ref = ccUKey2ref[refKey];
    if (!ref) return;

    if (ref.__$$ms === UNMOUNTED) {
      return;
    }

    var refCtx = ref.ctx;

    var _computeValueForRef = (0, _computeValueForRef2["default"])(ref, moduleName, prevModuleState, partialSharedState, callInfo, false, false),
        hasDeltaInCu = _computeValueForRef.hasDelta,
        cuCommittedState = _computeValueForRef.newCommittedState;

    var _watchKeyForRef = (0, _watchKeyForRef2["default"])(ref, moduleName, prevModuleState, partialSharedState, callInfo, false, false),
        hasDeltaInWa = _watchKeyForRef.hasDelta,
        waCommittedState = _watchKeyForRef.newCommittedState; // computed & watch 过程中提交了新的state，合并到 unProxyState 里
    // 注意这里，computeValueForRef watchKeyForRef 调用的 findDepFnsToExecute 内部
    // 保证了实例里cu或者wa函数 commit 提交的状态只能是 privateStateKey，所以合并到 unProxyState 是安全的


    if (hasDeltaInCu || hasDeltaInWa) {
      var changedRefPrivState = Object.assign(cuCommittedState, waCommittedState);
      var refModule = refCtx.module;
      var refState = refCtx.unProxyState;
      (0, _computeValueForRef2["default"])(ref, refModule, refState, changedRefPrivState, callInfo);
      (0, _watchKeyForRef2["default"])(ref, refModule, refState, changedRefPrivState, callInfo);
      Object.assign(refState, changedRefPrivState);
      Object.assign(refCtx.state, changedRefPrivState);

      refCtx.__$$settedList.push({
        module: refModule,
        keys: okeys(changedRefPrivState)
      });
    } // 记录 sharedStateKeys，方便 triggerRefEffect 之用


    refCtx.__$$settedList.push({
      module: moduleName,
      keys: sharedStateKeys
    });

    var upCb = function upCb() {
      return refCtx.__$$ccForceUpdate();
    };

    if (ref.__$$ms === NOT_MOUNT) {
      refCtx.__$$queuedUpdaters.push(upCb);
    } else {
      queuedOrUpdate(refKey, upCb);
    }
  });
  setTimeout(function () {
    updaterQueue.forEach(function (updater) {
      return updater();
    });
  }, 0);
}

function startChangeRefState(state, options, ref) {
  /**
   * 避免死循环，利用 setTimeout 将执行流程放到下一轮事件循环里
   *  在 <= v2.10.13之前
   *  1 watch 回调里执行 setState 导致无限死循环
   *  2 setup 块里直接执行 setState 导致无限死循环
   * 
   *  以 watch 为例：
   * function setup({watch, setState, initState}){
   *   initState({privKey: 2});
   *   watch('num', ()=>{
   *     // 因为watch是在组件渲染前执行，当设置 immediate 为 true 时
   *     // 组件处于 beforeMount 步骤，cUKey2Ref 并未记录具体的 ref,
   *     // 此时回调里调用setState会导致 use-concent 134 [KEY_1] 处判断失败后
   *     // 然后一直触发 cref 函数，一直进入新的 beforeMount 流程
   *     setState({privKey:1});
   *   }, {immediate:true});
   * }
   */
  if (ref.ctx.__$$inBM) {
    // <= 2.15.7
    // setTimeout(() => startChangeRefState(state, options, ref), 0);
    // > 2.15.7 调整为此逻辑
    // 满足一些的确需要在 setup 里及时的将数据写入 store 的场景
    // 由 permanentDispatcher 去触发其他组件实例渲染
    // 自身的 state 直接合入，这样在实例首次渲染的函数体能拿到 setup 里写入的最新状态
    // >=2.21.1 在broadcastState 内部新增延迟到下一轮事件循环更新的机制
    var permanentDispatcher = _ccContext["default"].getDispatcher();

    if (permanentDispatcher) {
      permanentDispatcher.ctx.changeState(state, options);
    }

    Object.assign(ref.ctx.state, state);
    Object.assign(ref.state, state);
    return;
  }

  changeRefState(state, options, ref);
}