"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

exports.__esModule = true;
exports.bindToContainer = bindToContainer;
exports.ccClassDisplayName = ccClassDisplayName;
exports.checkDepKeys = checkDepKeys;
exports.clearObject = clearObject;
exports.color = color;
exports.computeFeature = computeFeature;
exports.convertToStandardEvent = convertToStandardEvent;
exports.delay = delay;
exports.differStateKeys = differStateKeys;
exports.evalState = evalState;
exports.excludeArrStringItem = excludeArrStringItem;
exports.extractChangedState = extractChangedState;
exports.extractRenderKey = extractRenderKey;
exports.getErrStackKeywordLoc = getErrStackKeywordLoc;
exports.getRegisterOptions = getRegisterOptions;
exports.getVal = getVal;
exports.getValueByKeyPath = getValueByKeyPath;
exports.isArray = isArray;
exports.isAsyncFn = isAsyncFn;
exports.isBool = isBool;
exports.isDepKeysValid = isDepKeysValid;
exports.isEmptyVal = isEmptyVal;
exports.isFn = isFn;
exports.isKeyValid = isKeyValid;
exports.isLocal = isLocal;
exports.isModuleNameCcLike = isModuleNameCcLike;
exports.isModuleNameValid = isModuleNameValid;
exports.isObject = isObject;
exports.isObjectNotNull = isObjectNotNull;
exports.isObjectNull = isObjectNull;
exports.isOnlineEditor = isOnlineEditor;
exports.isPJO = isPJO;
exports.isSymbol = isSymbol;
exports.isValueNotNull = isValueNotNull;
exports.justTip = justTip;
exports.justWarning = justWarning;
exports.logErr = logErr;
exports.logNormal = logNormal;
exports.logWarn = logWarn;
exports.makeCallInfo = makeCallInfo;
exports.makeCcClassContext = makeCcClassContext;
exports.makeCommitHandler = makeCommitHandler;
exports.makeCuDepDesc = makeCuDepDesc;
exports.makeCuPackedValue = makeCuPackedValue;
exports.makeError = makeError;
exports.makeFnDesc = makeFnDesc;
exports.makeHandlerKey = makeHandlerKey;
exports.makeUniqueCcKey = makeUniqueCcKey;
exports.noDupPush = noDupPush;
exports.noop = noop;
exports.okeys = okeys;
exports.randomNumber = randomNumber;
exports.removeArrElements = removeArrElements;
exports.safeAdd = safeAdd;
exports.safeAssignObjectValue = safeAssignObjectValue;
exports.safeAssignToMap = safeAssignToMap;
exports.safeGet = safeGet;
exports.safeGetArray = safeGetArray;
exports.safeGetThenNoDupPush = safeGetThenNoDupPush;
exports.safeMinus = safeMinus;
exports.shallowCopy = shallowCopy;
exports.shallowDiffers = shallowDiffers;
exports.shouldGuessStrictMode = shouldGuessStrictMode;
exports.strictWarning = strictWarning;
exports.styleStr = styleStr;
exports.verboseInfo = verboseInfo;
exports.verifyKeys = verifyKeys;

var _extends2 = _interopRequireDefault(require("@babel/runtime/helpers/extends"));

var _constant = require("./constant");

var _privConstant = require("./priv-constant");

var _runtimeVar = _interopRequireDefault(require("../cc-context/runtime-var"));

/* eslint-disable */
var cer = function cer() {
  return _runtimeVar["default"].log && logErr.apply(void 0, arguments);
};

var protoToString = Object.prototype.toString;

function logErr() {
  var _console;

  (_console = console).error.apply(_console, arguments);
}

function logWarn() {
  var _console2;

  (_console2 = console).warn.apply(_console2, arguments);
}

function logNormal() {
  var _console3;

  (_console3 = console).log.apply(_console3, arguments);
}

function noop() {}

function isLocal() {
  return window && window.location && window.location.port;
}

function isValueNotNull(value) {
  return !(value === null || value === undefined);
}

function isObjectNotNull(object) {
  if (object === null || object === undefined) {
    return false;
  }

  if (okeys(object).length > 0) {
    return true;
  }

  return false;
}

function isObjectNull(object) {
  return !isObjectNotNull(object);
}

function isBool(val) {
  return typeof val === 'boolean';
} // !!!编译后的对象可能重写了toStringTag Symbol(Symbol.toStringTag): "Module"


var objStrList = ['[object Object]', '[object Module]', '[object Map]', '[object Set]'];

function isObject(obj) {
  if (!obj) return false; // undefined null etc...

  var str = protoToString.call(obj);
  return objStrList.includes(str);
}

function isArray(obj) {
  return Array.isArray(obj);
} // isPJO is short of isPlainJsonObject


function isPJO(obj, canBeArray) {
  if (canBeArray === void 0) {
    canBeArray = false;
  }

  var isArr = isArray(obj);
  var isObj = isObject(obj);
  return canBeArray ? isArr || isObj : isObj;
}

function isFn(maybeFn) {
  return typeof maybeFn === _privConstant.FN;
}

function isAsyncFn(fn, asyncKey) {
  if (!fn) return false; // @see https://github.com/tj/co/blob/master/index.js
  // obj.constructor.name === 'AsyncFunction'

  var isAsync = protoToString.call(fn) === '[object AsyncFunction]' || isFn(fn.then);

  if (isAsync === true) {
    return true;
  } // 有可能成降级编译成 __awaiter格式的了 或者 _regenerator


  var fnStr = fn.toString();

  if (fnStr.indexOf('_awaiter') >= 0 || fnStr.indexOf('_regenerator') >= 0) {
    return true;
  }
  /**
   * 上面的判定过程目前对这种编译结果是无效的，
   * function asyncFn(_x, _x2, _x3) {
   *     return _asyncFn.apply(this, arguments);
   *  }
   * 所以要求用户传入相应的asyncKeys来辅助判断，由runOptins里传入
   */


  if (asyncKey && _runtimeVar["default"].asyncCuKeys.includes(asyncKey)) {
    return true;
  }

  return false;
} // 0 算有效值, undefined null ''算空值


function isEmptyVal(val) {
  return !val && val !== 0;
}

function isKeyValid(obj, key) {
  return typeof key !== "symbol" && Object.prototype.hasOwnProperty.call(obj, key);
} // renderKey 可能是 IDispatchOptions


function extractRenderKey(renderKey) {
  var getRkey = function getRkey(key) {
    if (!key && key !== 0) return [];
    if (isArray(key)) return key;
    return null;
  };

  var targetRenderKey = getRkey(renderKey);
  if (targetRenderKey) return targetRenderKey;
  if (typeof renderKey === 'object') targetRenderKey = getRkey(renderKey.renderKey);
  if (targetRenderKey) return targetRenderKey;
  return [renderKey]; // 是一个具体的string 或 number
}

function makeError(code, extraMessage) {
  var message = '';
  if (typeof code === 'string') message = code;else {
    message = _constant.ERR_MESSAGE[code] || '';
  }
  if (extraMessage) message += extraMessage;
  if (!message) message = "undefined message for code:" + code;
  var error = new Error(message);
  error.code = code;
  return error;
}

function makeCuPackedValue(isLazy, result, needCompute, fn, newState, oldState, fnCtx) {
  var _ref;

  return _ref = {}, _ref[_privConstant.CU_KEY] = 1, _ref.needCompute = needCompute, _ref.fn = fn, _ref.newState = newState, _ref.oldState = oldState, _ref.fnCtx = fnCtx, _ref.isLazy = isLazy, _ref.result = result, _ref;
}

function makeCuDepDesc() {
  return {
    retKey2fn: {},
    retKey2lazy: {},
    stateKey2retKeys: {},
    // 用于辅助依赖收集系统更新依赖之用，render逻辑书写 refCompute.*** moduleCompted.*** connectedCompute.yy.** 时触发
    retKey2stateKeys: {},
    fnCount: 0
  };
}
/** make ccClassContext */


function makeCcClassContext(module, ccClassKey, renderKeyClasses) {
  return {
    module: module,
    ccClassKey: ccClassKey,
    renderKeyClasses: renderKeyClasses
  };
} // !!! different ccClass enable own a same key


function makeUniqueCcKey(ccClassKey, featureStr) {
  return ccClassKey + "$" + featureStr;
}

function makeHandlerKey(ccUniqueKey, eventName, identity) {
  return ccUniqueKey + "$" + eventName + "$" + identity;
}

function isModuleNameValid(moduleName) {
  if (!moduleName) return false;
  return /^[\$\#\&a-zA-Z0-9_-]+$/.test(moduleName);
}

function isModuleNameCcLike(moduleName) {
  var name = moduleName.toLowerCase();
  return name.startsWith(_constant.MODULE_CC);
}

function verboseInfo(info) {
  return info ? " --verbose-info: " + info : '';
}

function ccClassDisplayName(className) {
  return "CC(" + className + ")";
}

function verifyKeys(keys1, keys2) {
  var duplicate = false,
      notArray = false,
      keyElementNotString = false;
  if (!isArray(keys1)) return {
    duplicate: duplicate,
    notArray: true,
    keyElementNotString: keyElementNotString
  };
  if (!isArray(keys2)) return {
    duplicate: duplicate,
    notArray: true,
    keyElementNotString: keyElementNotString
  };
  var len1 = keys1.length;
  var len2 = keys2.length;

  outLoop: for (var i = 0; i < len1; i++) {
    var tmpKey = keys1[i];

    if (typeof tmpKey !== 'string') {
      keyElementNotString = true;
      break outLoop;
    }

    for (var j = 0; j < len2; j++) {
      var tmpKey2 = keys2[j];

      if (typeof tmpKey2 !== 'string') {
        keyElementNotString = true;
        break outLoop;
      }

      if (tmpKey2 === tmpKey) {
        duplicate = true;
        break outLoop;
      }
    }
  }

  return {
    duplicate: duplicate,
    notArray: notArray,
    keyElementNotString: keyElementNotString
  };
}

function color(color) {
  if (color === void 0) {
    color = 'green';
  }

  return "color:" + color + ";border:1px solid " + color;
}

function styleStr(str) {
  return "%c" + str;
}

var tipStart = function tipStart(str) {
  return "------------ CC " + str + " ------------";
};

function justWarning(err) {
  cer(tipStart('WARNING'));

  if (err instanceof Error) {
    cer(err.message);
    cer(err.stack);
  } else {
    cer(err);
  }
}

function justTip(msg, tipColor) {
  if (tipColor === void 0) {
    tipColor = 'green';
  }

  if (!_runtimeVar["default"].log) return;
  console.log(tipStart('TIP'));
  console.log("%c" + msg, "color:" + tipColor + ";border:1px solid " + tipColor + ";");
}

function strictWarning(err) {
  cer(tipStart('WARNING'));
  cer(err);

  if (_runtimeVar["default"].isStrict) {
    throw err;
  }
}

function safeAdd(object, key, toAdd) {
  try {
    object[key] += toAdd;
  } catch (err) {
    object[key] = toAdd;
  }
}

function safeMinus(object, key, toMinus) {
  try {
    object[key] -= toMinus;
  } catch (err) {
    object[key] = 0;
  }
}

function safeGet(object, key, defaultVal) {
  if (defaultVal === void 0) {
    defaultVal = {};
  }

  var childrenObject = object[key];

  if (!childrenObject) {
    childrenObject = object[key] = defaultVal;
  }

  return childrenObject;
}

function safeGetArray(object, key, defaultVal) {
  if (defaultVal === void 0) {
    defaultVal = [];
  }

  return safeGet(object, key, defaultVal);
}

function noDupPush(arr, strItem) {
  if (!arr.includes(strItem)) arr.push(strItem);
}

function safeGetThenNoDupPush(object, key, strItem) {
  var arr = safeGetArray(object, key);
  noDupPush(arr, strItem);
}

function safeAssignObjectValue(assignTo, assignFrom) {
  okeys(assignFrom).forEach(function (key) {
    assignTo[key] = assignFrom[key];
  });
}
/**
 * 把某个object赋值到container[key]的map下，map存在就直接赋值，map不存在则先创建再赋值，确保map引用无变化
 * @param {*} container 对象容器
 * @param {*} key 字段名
 * @param {*} objectToBeenAssign 等待赋值的object
 */


function safeAssignToMap(container, key, objectToBeenAssign) {
  var map = container[key];
  if (!map) map = container[key] = {};
  safeAssignObjectValue(map, objectToBeenAssign);
}

function computeFeature(ccUniqueKey, state) {
  var stateKeys = okeys(state);
  var stateKeysStr = stateKeys.sort().join('|');
  return ccUniqueKey + "/" + stateKeysStr;
}

function randomNumber(lessThan) {
  if (lessThan === void 0) {
    lessThan = 52;
  }

  var seed = Math.random();
  return parseInt(seed * lessThan);
} // 在 object[key]存在且deepClear为true时，传入的reset会被忽略
// 传入deepClear是为了保持引用不变


function clearObject(object, excludeKeys, reset, deepClear) {
  if (excludeKeys === void 0) {
    excludeKeys = [];
  }

  if (deepClear === void 0) {
    deepClear = false;
  }

  if (isArray(object)) {
    var retainKeys = [];
    excludeKeys.forEach(function (key) {
      if (object.includes(key)) retainKeys.push(key);
    });
    object.length = 0;
    retainKeys.forEach(function (key) {
      return object.push(key);
    });
    return;
  }

  okeys(object).forEach(function (key) {
    if (excludeKeys.includes(key)) {
      // do nothing
      return;
    }

    var subMap = object[key];

    if (deepClear && subMap) {
      okeys(subMap).forEach(function (key) {
        return delete subMap[key];
      });
    } else {
      if (reset) object[key] = reset;else delete object[key];
    }
  });
}

function okeys(obj) {
  return Object.keys(obj);
}

function excludeArrStringItem(arr, toExcludeStr) {
  var idx = arr.indexOf(toExcludeStr);

  if (idx > -1) {
    var arrCopy = arr.slice();
    arrCopy.splice(idx, 1);
    return arrCopy;
  } else {
    return arr;
  }
}

function convertToStandardEvent(e) {
  var ret = null; // avoid Warning: This synthetic event is reused for performance reasons. If you're seeing this...
  // call e.persist() @see https://reactjs.org/docs/events.html#event-pooling

  if (e) {
    if (e.persist) e.persist();

    if (e.currentTarget && e.type) {
      ret = e;
    } else if (e.nativeEvent && e.target) {
      e.currentTarget = e.target;
      ret = e;
    }
  }

  return ret;
}

function bindToContainer(key, toBindObj, targetContainerObj) {
  if (targetContainerObj) targetContainerObj[key] = toBindObj;
}
/**
 * 浅比较两个对象
 * come from : https://github.com/developit/preact-compat/blob/7c5de00e7c85e2ffd011bf3af02899b63f699d3a/src/index.js#L349
 */


function shallowDiffers(a, b) {
  for (var i in a) {
    if (!(i in b)) return true;
  }

  for (var _i in b) {
    if (a[_i] !== b[_i]) return true;
  }

  return false;
}

function shallowCopy(oriVal) {
  var newVal = oriVal;

  if (isObject(oriVal)) {
    newVal = (0, _extends2["default"])({}, oriVal);
  } else if (isArray(oriVal)) {
    newVal = [].concat(oriVal);
  }

  return newVal;
}

function extractChangedState(oldState, partialNewState, moduleOpt, force) {
  var changedState = {};
  var setted = false;
  var extractRefChangedState = _runtimeVar["default"].extractRefChangedState,
      extractModuleChangedState = _runtimeVar["default"].extractModuleChangedState,
      nonObjectValueCompare = _runtimeVar["default"].nonObjectValueCompare,
      objectValueCompare = _runtimeVar["default"].objectValueCompare;
  var needExtractChangedState = moduleOpt ? extractModuleChangedState : extractRefChangedState; // 非模块调用

  if (!moduleOpt) {
    if (!needExtractChangedState) return partialNewState;
    if (!nonObjectValueCompare && !objectValueCompare) return partialNewState;
  }

  if (partialNewState) {
    okeys(partialNewState).forEach(function (key) {
      var oldVal = oldState[key];
      var newVal = partialNewState[key];
      var valType = typeof newVal;
      var isNotEqual = true;

      if (force === true) {// let isNotEqual always be true
        // pass
      } else {
        if (valType !== 'object') {
          // 比较非object类型的值
          if (nonObjectValueCompare) isNotEqual = oldVal !== newVal;
        } else {
          // 比较object类型的值
          if (objectValueCompare) isNotEqual = oldVal !== newVal;
        }
      }

      if (isNotEqual) {
        if (moduleOpt) {
          moduleOpt.prevStateContainer[key] = oldVal;
          moduleOpt.incStateVer(key);
          oldState[key] = newVal;
        }

        changedState[key] = newVal;
        setted = true;
      }
    });
  }

  return setted ? changedState : null;
}

function differStateKeys(oldState, newState) {
  var changed = [],
      setted = [];
  okeys(newState).forEach(function (k) {
    var newVal = newState[k];
    setted.push(k);
    if (newVal !== oldState[k]) changed.push(k);
  });
  return {
    changed: changed,
    setted: setted
  };
}

function removeArrElements(arr, toRemoveArr) {
  var newArr = [];
  arr.forEach(function (item) {
    if (!toRemoveArr.includes(item)) newArr.push(item);
  });
  return newArr;
}

function getRegisterOptions(options) {
  if (options === void 0) {
    options = {};
  }

  if (typeof options === 'string') {
    return {
      module: options
    };
  }

  if (options) {
    if (options.module) return Object.assign({
      module: _constant.MODULE_DEFAULT
    }, options);
    return Object.assign(options, {
      module: _constant.MODULE_DEFAULT
    });
  }

  return {
    module: _constant.MODULE_DEFAULT
  };
}

function makeCommitHandler() {
  var state = null;

  var commit = function commit(partialState) {
    if (!state) state = {};
    Object.assign(state, partialState);
  };

  var clear = function clear() {
    return state = null;
  };

  var getFnCommittedState = function getFnCommittedState() {
    return state;
  };

  return {
    commit: commit,
    clear: clear,
    getFnCommittedState: getFnCommittedState
  };
}

function isOnlineEditor() {
  var result = false;

  if (window) {
    if (window.name === 'previewFrame' // for stackblitz
    || window.__SANDBOX_DATA__ // for codesandbox
    || window.BrowserFS // for codesandbox
    ) {
      result = true;
    }
  }

  return result;
}

var cachedSGSMResult = null;

function shouldGuessStrictMode() {
  if (cachedSGSMResult === null) {
    cachedSGSMResult = isLocal() || isOnlineEditor();
  }

  return cachedSGSMResult;
}

function makeCallInfo(module) {
  return {
    payload: null,
    renderKey: [],
    delay: -1,
    module: module,
    fnName: ''
  };
}

function evalState(state, mark) {
  if (state === void 0) {
    state = {};
  }

  if (mark === void 0) {
    mark = '';
  }

  var ret = isFn(state) ? state() : state;

  if (!isPJO(ret)) {
    throw new Error(mark + "state " + _privConstant.INAJ);
  }

  return ret;
}

function _getValue(obj, keys, lastKeyIndex, keyIndex) {
  var key = keys[keyIndex];

  if (lastKeyIndex === keyIndex) {
    return obj[key];
  } else {
    return _getValue(obj[key], keys, lastKeyIndex, ++keyIndex);
  }
}

function getValueByKeyPath(obj, keyPath) {
  var keys = keyPath.split('.');
  return _getValue(obj, keys, keys.length - 1, 0);
}

function isDepKeysValid(depKeys) {
  return isFn(depKeys) || isArray(depKeys) || depKeys === '-' || depKeys === '*';
}

function checkDepKeys(depKeys) {
  if (depKeys && !isDepKeysValid(depKeys)) {
    throw new Error("depKeys must be one of them(array,'*','-',fn)");
  }
}

function makeFnDesc(fn, depKeysOrOpt, check) {
  if (check === void 0) {
    check = true;
  }

  // 防止显式的传递null
  var _depKeysOrOpt = depKeysOrOpt || {};

  var desc = {
    fn: fn
  };
  var assignFrom = isDepKeysValid(_depKeysOrOpt) ? {
    depKeys: _depKeysOrOpt
  } : _depKeysOrOpt;
  check && checkDepKeys(assignFrom.depKeys);
  return Object.assign(desc, assignFrom);
}

var symbolTag = "[object Symbol]";

function isObjectLike(value) {
  return typeof value == "object" && value !== null;
}

function isSymbol(value) {
  return typeof value === "symbol" || isObjectLike(value) && Object.prototype.toString.call(value) === symbolTag;
}

function delay(ms) {
  if (ms === void 0) {
    ms = 1000;
  }

  return new Promise(function (resolve) {
    return setTimeout(resolve, ms);
  });
}

function getErrStackKeywordLoc(err, keyword, offset) {
  if (offset === void 0) {
    offset = 0;
  }

  var errStack = err.stack;
  var arr = errStack.split('\n');
  var len = arr.length;
  var curLocation = '';

  for (var i = 0; i < len; i++) {
    if (arr[i].includes(keyword)) {
      curLocation = arr[i + offset];
      break;
    }
  }

  return curLocation;
}

function getVal(val, defaultVal) {
  if (val !== undefined) return val;
  return defaultVal;
}